import asyncio

from pnq.selectors import starmap

from .._sync_generate.finalizers import (  # _all,; _any,; _sum,; each,; each,
    _len,
    _max,
    _min,
    accumulate,
    average,
    concat,
    contains,
    empty,
    exists,
    find,
    first,
    first_or,
    first_or_raise,
    last,
    last_or,
    last_or_raise,
    one,
    one_or,
    one_or_raise,
    reduce,
    to,
)
from ..common import Listable, name_as


@name_as("all")
def _all(source, selector=None) -> bool:
    return all(Listable(source, selector))


@name_as("any")
def _any(source, selector=None) -> bool:
    return any(Listable(source, selector))


@name_as("sum")
def _sum(source, selector=None):
    return sum(Listable(source, selector))


def each(source, func=lambda x: x, unpack=""):
    if asyncio.iscoroutinefunction(func):
        raise TypeError(f"cannot acceptable async funcition: {func}")

    func = starmap(func, unpack)

    for elm in source:
        func(elm)



import asyncio
from asyncio import iscoroutinefunction
from collections import deque
from concurrent.futures import Future
from functools import partial
from typing import Iterable, TypeVar

from pnq import selectors
from pnq.concurrent import get_default_pool
from pnq.protocols import PExecutor

from .queries import chunk

T = TypeVar("T")


def _procceed(func, iterable):
    return [func(x) for x in iterable]


async def _procceed_async(func, iterable):
    return [await func(x) for x in iterable]


def parallel(source: Iterable[T], func, executor: PExecutor, *, unpack="", chunksize=1):
    if executor is None:
        executor = get_default_pool()
    new_func = selectors.starmap(func, unpack)
    submit = executor.submit

    if executor.is_cpubound and chunksize != 1:
        runner = _procceed_async if asyncio.iscoroutine(func) else _procceed
        runner = partial(runner, new_func)

        tasks = [submit(runner, chunked) for chunked in chunk(source, chunksize)]
        for task in tasks:
            for x in task.result():
                yield x

    else:
        tasks = [submit(new_func, x) for x in source]
        for task in tasks:
            yield task.result()


def dispatch(
    source: Iterable[T],
    func,
    executor: PExecutor = None,
    *,
    unpack="",
    chunksize=1,
    on_complete=None,
):
    if executor is None:
        executor = get_default_pool()

    new_func = selectors.starmap(func, unpack)
    submit = executor.submit

    if on_complete is None:
        on_complete = lambda x: x  # noqa

    if executor.is_cpubound and chunksize != 1:
        runner = _procceed_async if asyncio.iscoroutine(func) else _procceed
        runner = partial(runner, new_func)

        for chunked in chunk(source, chunksize):
            future = submit(runner, chunked)
            future.add_done_callback(on_complete)

    else:
        for x in source:
            future = submit(new_func, x)
            future.add_done_callback(on_complete)


def exec_request(func, *args, **kwargs):
    from ..requests import Response, StopWatch

    with StopWatch() as sw:
        err = None
        result = None
        try:
            result = func(*args, **kwargs)
        except Exception as e:
            err = e

    return Response(
        func,
        args,
        kwargs,
        err=err,
        res=result,
        start=sw.start,
        end=sw.end,
    )


async def exec_request_async(func, *args, **kwargs):
    from ..requests import Response, StopWatch

    with StopWatch() as sw:
        err = None
        result = None
        try:
            result = await func(*args, **kwargs)
        except Exception as e:
            err = e

    return Response(
        func,
        args,
        kwargs,
        err=err,
        res=result,
        start=sw.start,
        end=sw.end,
    )


def request(
    source: Iterable[T],
    func,
    executor: PExecutor,
    *,
    unpack="",
    chunksize=1,
    retry: int = None,
    timeout: float = None,
):
    if executor is None:
        executor = get_default_pool()

    if iscoroutinefunction(func):
        wrapped = partial(exec_request_async, func)
    else:
        wrapped = partial(exec_request, func)

    return parallel(source, wrapped, executor, unpack=unpack, chunksize=chunksize)


def gather(
    source: Iterable[T], parallel: int = 1, timeout=None, return_exceptions=True
):
    import time

    tasks = {}
    queue = deque()  # type: ignore

    def set_result(i, future):
        del tasks[i]
        queue.append(future)

    for i, concurrent_future in enumerate(source):
        if isinstance(concurrent_future, Future):
            tasks[i] = concurrent_future
            concurrent_future.add_done_callback(partial(set_result, i))
        else:
            yield concurrent_future.result()  # type: ignore

        while queue:
            yield queue.popleft().result()

    while tasks:
        while queue:
            yield queue.popleft().result()
        time.sleep(0.01)

    while queue:
        yield queue.popleft().result()



import requests_unixsocket

from pnq import query

# https://docs.docker.com/registry/spec/api/


# IF ERROR
"""
{
    "errors:" [{
            "code": <error identifier>,
            "message": <message describing condition>,
            "detail": <unstructured>
        },
        ...
    ]
}
"""


class DockerClient:
    def __init__(
        self, schema="http+unix://%2Fvar%2Frun%2Fdocker.sock", version="v1.40"
    ):
        self.schema = schema
        self.version = version

    def _make_url(self, endpoint):
        schema = self.schema
        version = f"/{self.version}"
        return schema + version + endpoint

    def commands(self):
        return query(
            [
                "images",
                "containers",
                "networks",
                "volumes",
                "nodes",
                "services",
                "tasks",
                "secrets",
                "configs",
            ]
        )

    def images(self):
        url = self._make_url("/images/json")
        return query(requests_unixsocket.get(url).json())

    def containers(self):
        url = self._make_url("/containers/json")
        return query(requests_unixsocket.get(url).json())

    def networks(self):
        url = self._make_url("/networks/json")
        return query(requests_unixsocket.get(url).json())

    def volumes(self):
        url = self._make_url("/volumes/json")
        return query(requests_unixsocket.get(url).json())

    # swarm
    def nodes(self):
        url = self._make_url("/nodes/json")
        return query(requests_unixsocket.get(url).json())

    def services(self):
        url = self._make_url("/services/json")
        return query(requests_unixsocket.get(url).json())

    def tasks(self):
        url = self._make_url("/tasks/json")
        return query(requests_unixsocket.get(url).json())

    def secrets(self):
        url = self._make_url("/secrets/json")
        return query(requests_unixsocket.get(url).json())

    def configs(self):
        url = self._make_url("/configs/json")
        return query(requests_unixsocket.get(url).json())

    def plugins(self):
        url = self._make_url("/plugins/json")
        return query(requests_unixsocket.get(url).json())


Client = DockerClient



import platform
from glob import iglob
from os.path import isfile
from pathlib import Path

from pnq import query

# TODO: 3.10以降では削除
V_3_10 = int(platform.python_version_tuple()[1]) >= 10


# TODO: drop from 3.10
def _build_args(**kwargs):
    if V_3_10:
        return kwargs
    else:
        del kwargs["root_dir"]
        del kwargs["dir_fd"]
        return kwargs


def pathlib_to_str(x):
    if isinstance(x, Path):
        return str(x)
    else:
        return x


def ls(pathname="*", *, root_dir=None, dir_fd=None, recursive=False):
    pathname = pathlib_to_str(pathname)
    kwargs = _build_args(
        pathname=pathname, root_dir=root_dir, dir_fd=dir_fd, recursive=recursive
    )
    return query(Path(x) for x in iglob(**kwargs))


def files(pathname="*", *, root_dir=None, dir_fd=None, recursive=False):
    pathname = pathlib_to_str(pathname)
    kwargs = _build_args(
        pathname=pathname, root_dir=root_dir, dir_fd=dir_fd, recursive=recursive
    )
    return query(Path(x) for x in iglob(**kwargs) if isfile(x))


def dirs(pathname="*", *, root_dir=None, dir_fd=None, recursive=False):
    pathname = pathlib_to_str(pathname)
    kwargs = _build_args(
        pathname=pathname, root_dir=root_dir, dir_fd=dir_fd, recursive=recursive
    )
    return query(Path(x) for x in iglob(**kwargs) if not isfile(x))



from typing import Any, Callable

__all__ = ["item", "attr"]


def _get_getter(is_attr):
    if is_attr:
        return getattr
    else:
        return lambda x, k: x[k]


def eq(field, value, attr=False):
    getter = _get_getter(attr)
    return lambda x: getter(x, field) == value


def ne(field, value, attr=False):
    getter = _get_getter(attr)
    return lambda x: getter(x, field) != value


class Item:
    def __getattribute__(self, name: str) -> "ItemOperator":
        return ItemOperator(name)


class Attr:
    def __getattribute__(self, name: str) -> "AttrOperator":
        return AttrOperator(name)


class ItemOperator:
    def __init__(self, key) -> None:
        self.key = key

    def __eq__(self, o: object) -> Callable[[Any], bool]:  # type: ignore
        return eq(self.key, o, attr=False)

    def __ne__(self, o: object) -> Callable[[Any], bool]:  # type: ignore
        return ne(self.key, o, attr=False)


class AttrOperator:
    def __init__(self, key) -> None:
        self.key = key

    def __eq__(self, o: object) -> Callable[[Any], bool]:  # type: ignore
        return eq(self.key, o, attr=True)

    def __ne__(self, o: object) -> Callable[[Any], bool]:  # type: ignore
        return ne(self.key, o, attr=True)


item = Item()
attr = Attr()



from typing import AsyncIterable, Callable, Generic, Iterable, TypeVar, Union

T = TypeVar("T")


class Listable(Iterable[T]):
    def __init__(self, iterable: Iterable[T]):
        self.source = iterable

    def __iter__(self):
        return iter(self.source)

    def save(self):
        return self.result()

    def result(self, timeout=None):
        # TODO: implement timeout
        return list(self.source)

    def __future__(self):
        raise NotImplementedError()


class AsyncListable(AsyncIterable[T]):
    def __init__(self, iterable: AsyncIterable[T]):
        self.source = iterable

    def __aiter__(self):
        return self.source.__aiter__()

    async def save_async(self):
        return [x async for x in self]

    def __await__(self):
        return self.save_async().__await__()

    def __afuture__(self):
        raise NotImplementedError()


class LazyListable(Listable[T], AsyncListable[T]):  # type: ignore
    def __init__(self, iterable: Union[Iterable[T], AsyncIterable[T]]):
        self.source = iterable  # type: ignore

    def as_iter(self) -> Listable[T]:
        return Listable(self.source)

    def as_aiter(self) -> AsyncListable[T]:
        return AsyncListable(self.source)  # type: ignore


class LazyGeneratorFunction(Generic[T]):
    def __init__(
        self, func: Union[Callable[..., Iterable[T]], Callable[..., AsyncIterable[T]]]
    ):
        self.func = func

    def __call__(self, *args, **kwargs) -> LazyListable[T]:
        return LazyListable(self.func(*args, **kwargs))



import time
from itertools import product as _product

from .._sync_generate.queries import *  # noqa
from .._sync_generate.queries import _enumerate, _filter, _map, _take_page_calc

_zip = zip
# _enumerate = enumerate


def order_by_reverse(source):  # type: ignore
    if hasattr(source, "__reversed__"):
        return reversed(source)
    else:

        def reverse_iterator():
            yield from reversed(list(source))

        return reverse_iterator()


def sleep(source, seconds: float):  # type: ignore
    sleep = time.sleep
    for v in source:
        yield v
        sleep(seconds)


cartesian = _product


def chain(*iterables):  # type: ignore
    for iterable in iterables:
        for x in iterable:
            yield x



import asyncio
from asyncio import iscoroutinefunction
from concurrent.futures import Future as ConcurrentFuture
from functools import cmp_to_key
from functools import partial
from functools import partial as _partial
from operator import attrgetter, itemgetter
from typing import Any, Awaitable, NoReturn

from typing_extensions import Literal

__all__ = [
    "starmap",
    "select",
    "select_from_attr",
    "select_from_item",
    "select_as_tuple",
    "select_as_dict",
    "select_as_awaitable",
    "select_as_future",
    "reflect",
    "traverse",
    "cmp_to_key",
]


def to_str(x):
    return "" if x is None else str(x)


def _star1map(func, val):
    return func(*val)


def _star2map(func, val):
    return func(**val)


def _star3map(func, val):
    return func(*val.args, **val.kwargs)


async def _star1map_async(func, val):
    return await func(*val)


async def _star2map_async(func, val):
    return await func(**val)


async def _star3map_async(func, val):
    return await func(*val.args, **val.kwargs)


def star1map(func):
    if iscoroutinefunction(func):
        return _partial(_star1map_async, func)
    else:
        return _partial(_star1map, func)


def star2map(func):
    if iscoroutinefunction(func):
        return _partial(_star2map_async, func)
    else:
        return _partial(_star2map, func)


def star3map(func):
    if iscoroutinefunction(func):
        return _partial(_star3map_async, func)
    else:
        return _partial(_star3map, func)


def starmap(func, unpack: Literal["", "*", "**", "***"] = "*"):
    if func is None:
        raise TypeError("func is None")

    func = to_str if func is str else func

    if unpack == "":
        return func
    elif unpack == "*":
        return star1map(func)
    elif unpack == "**":
        return star2map(func)
    elif unpack == "***":
        return star3map(func)
    else:
        raise ValueError(f"Unsupported unpack mode: {unpack}")


def select_from_attr(*args):
    return attrgetter(*args)


def select_from_item(*args):
    return itemgetter(*args)


def select(*args, attr=False):
    if attr:
        return select_from_attr(*args)
    else:
        return select_from_item(*args)


def select_as_tuple(*args, attr=False):
    if len(args) == 0:
        return lambda x: ()

    if attr:
        getter = select_from_attr(*args)
    else:
        getter = select_from_item(*args)

    if len(args) == 1:
        return lambda x: (getter(x),)
    else:
        return getter


def select_as_dict(*fields, attr=False, default=NoReturn):
    if attr:
        if default is NoReturn:
            selector = lambda x: {k: getattr(x, k) for k in fields}  # noqa
        else:
            selector = lambda x: {k: getattr(x, k, default) for k in fields}  # noqa
    else:
        if default is NoReturn:
            selector = lambda x: {k: x[k] for k in fields}  # noqa
        else:

            def getitem(obj, k):
                try:
                    return obj[k]
                except Exception:
                    return default

            selector = lambda x: {k: getitem(x, k) for k in fields}  # noqa

    return selector


def reflect(mapping, attr: bool = False):
    transposed = _transpose(mapping)
    single, multi = _split_single_multi(transposed)
    return _build_selector(single, multi, attr)


def _transpose(mapping):
    from collections import defaultdict

    tmp = defaultdict(list)

    for left, right in mapping.items():
        if isinstance(right, str):
            tmp[left].append(right)
        elif isinstance(right, list):
            tmp[left] = right
        elif isinstance(right, tuple):
            tmp[left] = right
        elif isinstance(right, set):
            tmp[left] = right
        else:
            raise TypeError(f"{v} is not a valid mapping")

    # output属性 - 元の属性（複数の場合あり）
    target = defaultdict(list)

    for k, outputs in tmp.items():
        for out in outputs:
            target[out].append(k)

    return target


def _split_single_multi(dic):
    single = {}
    multi = {}
    for k, v in dic.items():
        if len(v) > 1:
            multi[k] = v
        else:
            single[k] = v[0]

    return single, multi


def _build_selector(single, multi, attr: bool = False):
    if attr:
        getter = getattr

    else:
        getter = dict.__getitem__

    def reflector(x):
        result = {}
        for k, v in single.items():
            result[k] = getter(x, v)

        for k, fields in multi.items():
            result[k] = []
            for f in fields:
                result[k].append(getter(x, f))

        return result

    return reflector


def traverse(selector):
    from collections import deque

    # breadth-first-search
    def traverse(root):
        queue = deque()  # type: ignore
        queue.append(root)
        while queue:
            node = queue.popleft()
            yield node
            children = selector(node)
            if children is None:
                children = []

            for x in children:
                queue.append(x)

    return traverse


async def _awaitable_wrapper(awaitable):
    return await awaitable


def _select_future_with_schedule(target) -> asyncio.Future:
    if hasattr(target, "__await__"):
        if asyncio.isfuture(target):
            return target
        elif asyncio.iscoroutine(target):
            return asyncio.create_task(target)
        else:
            aw = _awaitable_wrapper(target)
            return asyncio.create_task(aw)

    if isinstance(target, ConcurrentFuture):
        return asyncio.wrap_future(target)

    raise TypeError(f"{target} is not awaitable")



import asyncio
import concurrent.futures
from enum import Flag
from typing import (
    TYPE_CHECKING,
    Any,
    AsyncIterable,
    Awaitable,
    Generator,
    Generic,
    Iterable,
    List,
    TypeVar,
)

from typing_extensions import Protocol

T = TypeVar("T")


class PExecutor(Protocol):
    def submit(self, func, *args, **kwargs) -> concurrent.futures.Future:
        ...

    def asubmit(self, func, *args, **kwargs) -> asyncio.Future:
        ...

    @property
    def is_cpubound(self) -> bool:
        ...

    @property
    def max_workers(self) -> int:
        ...

    @property
    def is_async_only(self) -> bool:
        ...


class PResult(Protocol[T]):
    def result(self, timeout=None) -> List[T]:
        raise NotImplementedError()


class PAsyncResult(Awaitable[List[T]], Generic[T]):
    def __await__(self) -> Generator[Any, Any, List[T]]:
        raise NotImplementedError()


class IterType(Flag):
    IMPOSSIBLE = 0
    NORMAL = 1
    ASYNC = 2
    BOTH = NORMAL | ASYNC


class PQuery(Iterable[T], AsyncIterable[T], PResult[T], PAsyncResult[T], Generic[T]):
    iter_type: IterType



from dataclasses import dataclass
from functools import partial
from typing import Tuple


class Undefined:
    def __eq__(self, o):
        raise AttributeError("undefined cannot be referenced.")

    def __ne__(self, o):
        raise AttributeError("undefined cannot be referenced.")

    def __bool__(self):
        raise AttributeError("undefined cannot be referenced.")


undefined: bool = Undefined()  # type: ignore


class TestException(Exception):
    ...


@dataclass
class ExecutorSpec:
    is_cpubound: bool = undefined
    max_workers: bool = undefined
    is_async_only: bool = undefined
    __enter__: bool = undefined
    __exit__: bool = undefined
    __aenter__: bool = undefined
    __aexit__: bool = undefined
    submit_sync: bool = undefined
    submit_async: bool = undefined
    asubmit_sync: bool = undefined
    asubmit_async: bool = undefined
    submit_partial_sync: bool = undefined
    submit_partial_async: bool = undefined
    asubmit_partial_sync: bool = undefined
    asubmit_partial_async: bool = undefined

    @classmethod
    async def test_capability(
        cls, factory, is_async_only, submit_sync=True, submit_async=True
    ):
        _is_async_only, spec = await test_executor_spec(factory)
        # required
        assert spec.is_cpubound
        assert spec.max_workers
        assert spec.is_async_only
        assert spec.__aenter__
        assert spec.__aexit__
        assert _is_async_only == is_async_only

        if _is_async_only:
            assert not spec.__enter__
            assert not spec.__exit__
            assert spec.asubmit_sync
            assert spec.asubmit_async
            if not submit_sync:
                assert isinstance(spec.submit_sync, NotImplementedError)
                assert isinstance(spec.submit_partial_sync, NotImplementedError)
            else:
                assert spec.submit_sync
                assert spec.submit_partial_sync

            if not submit_async:
                assert isinstance(spec.submit_async, NotImplementedError)
                assert isinstance(spec.submit_partial_async, NotImplementedError)
            else:
                assert spec.submit_async
                assert spec.submit_partial_async

        else:
            assert submit_sync
            assert spec.__enter__
            assert spec.__exit__
            assert spec.submit_sync
            assert spec.submit_async
            assert spec.asubmit_sync
            assert spec.asubmit_async
            assert spec.submit_partial_sync
            assert spec.submit_partial_async
            assert spec.asubmit_partial_sync
            assert spec.asubmit_partial_async

        return True


def return_one():
    return 1


async def return_two():
    return 2


async def test_executor_spec(factory) -> Tuple[bool, ExecutorSpec]:
    spec = ExecutorSpec()

    try:
        with factory(1) as f:
            raise TestException()

    except TestException:
        spec.__enter__ = True
        spec.__exit__ = True

    except NotImplementedError:
        spec.__enter__ = False
        spec.__exit__ = False

    try:
        async with factory(1) as f:
            raise TestException()

    except TestException:
        spec.__aenter__ = True
        spec.__aexit__ = True

    async with factory(1) as executor:
        spec.is_cpubound = isinstance(executor.is_cpubound, bool)
        spec.max_workers = isinstance(executor.is_cpubound, int)
        spec.is_async_only = isinstance(executor.is_async_only, bool)
        spec.submit_sync = try_submit(executor, return_one, 1)
        spec.submit_async = try_submit(executor, return_two, 2)
        spec.asubmit_sync = await try_asubmit(executor, return_one, 1)
        spec.asubmit_async = await try_asubmit(executor, return_two, 2)

        # python3.8からはpartialはasync functionを認識できる模様
        spec.submit_partial_sync = try_submit(executor, partial(return_one), 1)
        spec.submit_partial_async = try_submit(executor, partial(return_two), 2)
        spec.asubmit_partial_sync = await try_asubmit(executor, partial(return_one), 1)
        spec.asubmit_partial_async = await try_asubmit(executor, partial(return_two), 2)

        return executor.is_async_only, spec


def try_submit(executor, func, expect):
    try:
        return executor.submit(func).result() == expect
    except Exception as e:
        return e


async def try_asubmit(executor, func, expect):
    try:
        return await executor.asubmit(func) == expect
    except Exception as e:
        return e



import os
from pathlib import Path

import unasync

# additional_replacements = {
#     # We want to rewrite to 'Transport' instead of 'SyncTransport', etc
#     "AsyncTransport": "Transport",
#     "AsyncElasticsearch": "Elasticsearch",
#     # We don't want to rewrite this class
#     "AsyncSearchClient": "AsyncSearchClient",
# }

PACKAGE = "pnq"


CODE_ROOT = Path(__file__).absolute().parent

rules = [
    unasync.Rule(
        fromdir=str(CODE_ROOT / "_async"),
        todir=str(CODE_ROOT / "_sync_generate"),
        # additional_replacements={"Listable": "map_iter", "Listable": "map_iter"},
    ),
]

filepaths = []
for root, _, filenames in os.walk(CODE_ROOT / "_async"):
    for filename in filenames:
        if filename.rpartition(".")[-1] in (
            "py",
            "pyi",
        ) and not filename.startswith("utils.py"):
            filepaths.append(os.path.join(root, filename))

# create _sync from _async
unasync.unasync_files(filepaths, rules)



from typing import AsyncIterable, Callable, Iterable, TypeVar, Union, overload

from . import iterables

T = TypeVar("T")


@overload
def LazyGeneratorFunction(
    func: Union[Callable[..., Iterable[T]], Callable[..., AsyncIterable[T]]]
) -> iterables.LazyGeneratorFunction[T]:
    ...


@overload
def LazyGeneratorFunction(func) -> iterables.LazyGeneratorFunction:
    ...


def LazyGeneratorFunction(func):
    return iterables.LazyGeneratorFunction(func)


@overload
def LazyListable(func: AsyncIterable[T]) -> iterables.AsyncListable[T]:  # type: ignore
    ...


@overload
def LazyListable(func: Iterable[T]) -> iterables.Listable[T]:  # type: ignore
    ...


@overload
def LazyListable(func) -> iterables.LazyListable:
    ...


def LazyListable(func):
    return iterables.LazyListable(func)



from asyncio import sleep as asleep
from datetime import datetime
from time import sleep

from pnq import query
from pnq._itertools.requests import CancelToken


# TODO: 長いsecondsが与えられた時、その待機中トークンの判定がされない。sleepを細切れにして、できるだリアルタイムに判定されるようにする
def tick(seconds: float, token: CancelToken = None):
    token = token or CancelToken()

    def infinity(seconds, token: CancelToken):
        while token.is_active:
            yield datetime.utcnow()
            sleep(seconds)

    return query(infinity(seconds, token))


def tick_async(seconds: float, token: CancelToken = None):
    token = token or CancelToken()

    async def infinity(seconds, token: CancelToken):
        while token.is_active:
            yield datetime.utcnow()
            await asleep(seconds)

    return query(infinity(seconds, token))


class timezones:
    @classmethod
    def keys(cls):
        import zoneinfo

        return query(zoneinfo.available_timezones())

    @classmethod
    def values(cls):
        from zoneinfo import ZoneInfo

        return query(ZoneInfo(x) for x in cls.keys())

    @classmethod
    def items(cls):
        from zoneinfo import ZoneInfo

        return query((x, ZoneInfo(x)) for x in cls.keys())


class weekdays:
    """datetime.weekday()で返される値の対応表"""

    @classmethod
    def keys(cls):
        return query(x[0] for x in cls.items())

    @classmethod
    def values(cls):
        return query(x[1] for x in cls.items())

    @classmethod
    def items(cls):
        return query(
            {
                0: "Monday",
                1: "Tuesday",
                2: "Wednesday",
                3: "Thursday",
                4: "Friday",
                5: "Saturday",
                6: "Sunday",
            }
        )


class isoweekdays(weekdays):
    """datetime.isoweekday()で返される値の対応表"""

    @classmethod
    def items(cls):
        return query(
            {
                1: "Monday",
                2: "Tuesday",
                3: "Wednesday",
                4: "Thursday",
                5: "Friday",
                6: "Saturday",
                7: "Sunday",
            }
        )


class spans:
    @classmethod
    def keys(cls):
        return query(x[0] for x in cls.items())

    @classmethod
    def values(cls):
        return query(x[1] for x in cls.items())

    @classmethod
    def items(cls):
        return query(
            {
                1: "year",
                2: "month",
                3: "day",
                4: "hour",
                5: "minute",
                6: "second",
                7: "microsecond",
            }
        )



"""
Copyright (c) 2021 sasano8

Released under the MIT license.
see https://opensource.org/licenses/MIT

The universal_write_open function from:
https://github.com/EntilZha/PyFunctional/blob/master/functional/io.py
https://github.com/EntilZha/PyFunctional/blob/master/functional/pipeline.py
"""

import builtins
import bz2
import gzip
import lzma


def universal_write_open(
    path,
    mode,
    buffering=-1,
    encoding=None,
    errors=None,
    newline=None,
    compresslevel=9,
    format=None,
    check=-1,
    preset=None,
    filters=None,
    compression=None,
):
    # pylint: disable=unexpected-keyword-arg,no-member
    if compression is None:
        return builtins.open(
            path,
            mode=mode,
            buffering=buffering,
            encoding=encoding,
            errors=errors,
            newline=newline,
        )
    elif compression in ("gz", "gzip"):
        return gzip.open(
            path,
            mode=mode,
            compresslevel=compresslevel,
            errors=errors,
            newline=newline,
            encoding=encoding,
        )
    elif compression in ("lzma", "xz"):
        return lzma.open(
            path,
            mode=mode,
            format=format,
            check=check,
            preset=preset,
            filters=filters,
            encoding=encoding,
            errors=errors,
            newline=newline,
        )
    elif compression == "bz2":
        return bz2.open(
            path,
            mode=mode,
            compresslevel=compresslevel,
            encoding=encoding,
            errors=errors,
            newline=newline,
        )
    else:
        raise ValueError(
            "compression must be None, gz, gzip, lzma, or xz and was {0}".format(
                compression
            )
        )



import asyncio
from typing import Any, Callable, Coroutine, TypeVar

__all__ = ["CancelToken", "run", "get_cancel_token"]


R = TypeVar("R")


class CancelToken:
    def __init__(self):
        self.is_cancelled = False

    def is_running(self):
        return not self.is_cancelled

    def cancel(self):
        self.is_cancelled = True


def run(
    func: Callable[..., Coroutine[Any, Any, R]],
    handle_signals={"SIGINT", "SIGTERM"},
) -> R:
    import asyncio
    import inspect
    import signal
    from functools import partial

    cancel_count = 0

    def handle_cancel(signame, task, token=None):
        if token is None:
            print(
                f"Cancel requested by {signame}. The task will be forcibly canceled."  # noqa
            )
            task.cancel()
        else:
            print(
                f"Cancel requested by {signame}. The task will be safely shut down. If you want to force cancel, please cancel 5 times."
            )
            token.cancel()

            nonlocal cancel_count
            cancel_count += 1

            if cancel_count >= 5:
                task.cancel()
            if cancel_count >= 10:
                exit(1)

    signature = inspect.signature(func)

    token = CancelToken()
    if len(signature.parameters) == 1:
        args = [token]
    else:
        args = []

    loop = asyncio.new_event_loop()
    set_cancel_token(loop, token)
    future = asyncio.shield(func(*args), loop=loop)

    for sig_name in handle_signals:
        sig = getattr(signal, sig_name)
        loop.add_signal_handler(sig, partial(handle_cancel, sig_name, future, *args))

    try:
        loop.run_until_complete(future)
    finally:
        loop.close()

    return future.result()


def set_cancel_token(loop, token):
    loop._cancel_token = token


def get_cancel_token():
    loop = asyncio.get_running_loop()

    token = getattr(loop, "_cancel_token", None)
    if token is None:
        raise RuntimeError("No cancel token found. please start by 'pnq.aio.run(func)'")

    return token


def cancel():
    token = get_cancel_token()
    token._is_cancelled = True



import asyncio
import concurrent
from typing import Dict, Iterable

from typing_extensions import Protocol, runtime_checkable  # from python3.8

"""
concurrent.futures.ProcessPoolExecutorはmultiprocessing.Poolのラッパーです。
https://stackoverflow.com/questions/38311431/concurrent-futures-processpoolexecutor-vs-multiprocessing-pool-pool
"""


@runtime_checkable
class Executor(Protocol):
    def submit(self, *args, **kwargs) -> concurrent.futures.Future:
        ...


class PExecutor(Protocol):
    def submit(self, func, *args, **kwargs) -> concurrent.futures.Future:
        ...

    def asubmit(self, func, *args, **kwargs) -> asyncio.Future:
        ...

    @property
    def running_task_count(self) -> int:
        ...

    @property
    def is_full(self) -> bool:
        ...

    @property
    def is_closed(self) -> bool:
        ...

    @property
    def max_workers(self) -> int:
        ...

    @property
    def is_cpubound(self) -> bool:
        ...


class PExecutable(Protocol):
    def __executor__(self) -> PExecutor:
        ...



import csv
import json
from functools import partial
from typing import Iterable, Mapping, Union

from typing_extensions import Literal

from ..io import universal_write_open
from .finalizers import first_or


async def to_file(
    source,
    path,
    delimiter: str = None,
    mode: Union[Literal["r", "w", "x", "a", "b", "t", "+"], str] = "wt",
    buffering=-1,
    encoding=None,
    errors=None,
    newline=None,
    compresslevel=9,
    format=None,
    check=-1,
    preset=None,
    filters=None,
    compression: Union[None, Literal["gz", "gzip", "lzma", "xz", "bz2"]] = None,
):
    """
    Saves the sequence to a file by executing str(self) which becomes str(self.to_list()). If
    delimiter is defined will instead execute self.make_string(delimiter)
    :param path: path to write file
    :param delimiter: if defined, will call make_string(delimiter) and save that to file.
    :param mode: file open mode
    :param buffering: passed to builtins.open
    :param encoding: passed to builtins.open
    :param errors: passed to builtins.open
    :param newline: passed to builtins.open
    :param compression: compression format
    :param compresslevel: passed to gzip.open
    :param format: passed to lzma.open
    :param check: passed to lzma.open
    :param preset: passed to lzma.open
    :param filters: passed to lzma.open
    """
    with universal_write_open(
        path,
        mode=mode,
        buffering=buffering,
        encoding=encoding,
        errors=errors,
        newline=newline,
        compression=compression,
        compresslevel=compresslevel,
        format=format,
        check=check,
        preset=preset,
        filters=filters,
    ) as output:
        if delimiter is None:
            bracket_start = "["
            bracket_end = "]"
            delimiter = ","
        else:
            bracket_start = ""
            bracket_end = ""
            delimiter = delimiter

        undefined = object()
        it = source.__aiter__()
        first = await first_or(it, undefined)

        if first is undefined:
            if bracket_start:
                output.write(bracket_start)
                output.write(bracket_end)
            return

        if bracket_start:
            output.write(bracket_start)

        output.write(str(first))

        async for row in it:
            output.write(delimiter + str(row))

        if bracket_end:
            output.write(bracket_end)


async def to_json(
    source,
    path,
    as_dict=False,
    mode: Union[Literal["r", "w", "x", "a", "b", "t", "+"], str] = "wt",
    compression: Union[None, Literal["gz", "gzip", "lzma", "xz", "bz2"]] = None,
    ensure_ascii=False,
    indent=2,
):
    """
    Saves the sequence to a json file. If root_array is True, then the sequence will be written
    to json with an array at the root. If it is False, then the sequence will be converted from
    a sequence of (Key, Value) pairs to a dictionary so that the json root is a dictionary.
    :param path: path to write file
    :param root_array: write json root as an array or dictionary
    :param mode: file open mode
    """
    with universal_write_open(path, mode=mode, compression=compression) as output:
        dumps = partial(json.dumps, ensure_ascii=ensure_ascii, indent=indent)
        s_indent = " " * indent

        if not as_dict:
            output.write("[")
        else:
            output.write("{")

        undefined = object()
        it = source.__aiter__()
        first_ = await first_or(it, undefined)

        if first_ is not undefined:
            if not as_dict:
                output.write("\n" + s_indent + dumps(first_))
                s_indent = ",\n" + s_indent

                async for x in it:
                    output.write(s_indent + dumps(x))
            else:
                k, v = first_
                output.write("\n" + s_indent + str(dumps(k)) + ": " + dumps(v))
                s_indent = ",\n" + s_indent

                async for k, v in it:
                    output.write(s_indent + str(dumps(k)) + ": " + dumps(v))

        if not as_dict:
            output.write("\n]\n")
        else:
            output.write("\n}\n")


async def to_jsonl(
    source,
    path,
    mode: Union[Literal["r", "w", "x", "a", "b", "t", "+"], str] = "wt",
    compression: Union[None, Literal["gz", "gzip", "lzma", "xz", "bz2"]] = None,
    ensure_ascii=False,
):
    """
    Saves the sequence to a jsonl file. Each element is mapped using json.dumps then written
    with a newline separating each element.
    :param path: path to write file
    :param mode: mode to write in, defaults to 'w' to overwrite contents
    :param compression: compression format
    :param ensure_ascii: ensure_ascii
    """
    dumps = partial(json.dumps, ensure_ascii=ensure_ascii)
    with universal_write_open(path, mode=mode, compression=compression) as output:
        async for x in source:
            output.write((dumps(x) + "\n"))


async def to_csv(
    source,
    path,
    mode: Union[Literal["r", "w", "x", "a", "b", "t", "+"], str] = "wt",
    header: Union[None, bool, Iterable[str]] = True,
    # columns: Iterable[str] = None,
    skip_first: bool = False,
    delimiter: str = ",",
    dialect="excel",
    compression: Union[None, Literal["gz", "gzip", "lzma", "xz", "bz2"]] = None,
    newline="",
    **fmtparams,
):
    """
    Saves the sequence to a csv file. Each element should be an iterable which will be expanded
    to the elements of each row.
    :param path: path to write file
    :param mode: file open mode
    :param dialect: passed to csv.writer
    :param fmtparams: passed to csv.writer
    """

    if "fieldnames" in fmtparams:
        raise ValueError("fieldnames is not supported. use header.")

    if "b" in mode:
        newline = None

    with universal_write_open(
        path, mode=mode, compression=compression, newline=newline
    ) as output:
        # csv_writer = csv.writer(output, dialect=dialect, **fmtparams)

        if header is None or isinstance(header, int):
            output_header = bool(header)
            columns = None
        else:
            output_header = True
            columns = header

        header_printed = False

        if output_header and skip_first and columns is None:
            raise ValueError(
                f"The behavior is undefined. header:{header} skip_first:{skip_first}"
            )

        if output_header and columns is not None:
            csv.writer(
                output, dialect=dialect, delimiter=delimiter, **fmtparams
            ).writerow(columns)
            header_printed = True

        undefined = object()
        it = source.__aiter__()
        first_ = await first_or(it, undefined)

        if first_ is undefined:
            return

        if isinstance(first_, Mapping):
            if columns is None:
                columns = list(first_.keys())
            else:
                ...

            writer = csv.DictWriter(
                output,
                fieldnames=columns,
                dialect=dialect,
                delimiter=delimiter,
                **fmtparams,
            )
            if output_header and not header_printed:
                writer.writeheader()

            if not skip_first:
                writer.writerow(first_)

        else:

            if columns is None:
                columns = first_
            else:
                ...

            writer = csv.writer(
                output,
                dialect=dialect,
                delimiter=delimiter,
                **fmtparams,
            )
            if not skip_first:
                writer.writerow(first_)

        async for row in it:
            writer.writerow(row)



import typing

from pnq import query

DEFAULT_IGNORE_KEYS = {
    "TYPE_CHECKING",
    "runtime_checkable",
    "get_type_hints",
    "get_args",
    "is_typeddict",
    "no_type_check_decorator",
    "final",
    "cast",
    "no_type_check",
    "get_origin",
    "overload",
    # "ForwardRef",
}


def default_ignore_keys():
    return query(DEFAULT_IGNORE_KEYS)


def keys(ignores: typing.Iterable[str] = DEFAULT_IGNORE_KEYS):
    _types = set(typing.__all__)
    for key in ignores:
        _types.discard(key)

    return query(_types)


def values(ignores: typing.Iterable[str] = DEFAULT_IGNORE_KEYS):
    return query(getattr(typing, x) for x in keys(ignores))


def items(ignores: typing.Iterable[str] = DEFAULT_IGNORE_KEYS):
    return query((x, getattr(typing, x)) for x in keys(ignores))



from enum import Flag
from typing import (
    TYPE_CHECKING,
    AsyncIterable,
    AsyncIterator,
    Generic,
    Iterable,
    Iterator,
    Mapping,
    Tuple,
    TypeVar,
)

from . import finalizers

if TYPE_CHECKING:
    # python3.7には含まれていない
    from typing import final
else:
    final = lambda x: x  # noqa

T = TypeVar("T")
K = TypeVar("K")
V = TypeVar("V")


class IterType(Flag):
    IMPOSSIBLE = 0
    NORMAL = 1
    ASYNC = 2
    BOTH = NORMAL | ASYNC


def get_iter_type(source):
    run_iter_type = getattr(source, "run_iter_type", None)
    if run_iter_type:
        return run_iter_type
    elif hasattr(source, "__aiter__"):
        if hasattr(source, "__iter__"):
            return IterType.BOTH
        else:
            return IterType.ASYNC
    elif hasattr(source, "__iter__"):
        return IterType.NORMAL
    else:
        raise TypeError(f"{source} has no __iter__ or __aiter__")


class Query(Iterable[T], AsyncIterable[T], Generic[T]):
    """Queryクラスをチェインするのに使うか、__iter__と__aiter__の挙動をソースに任せる場合に使います。"""

    iter_type = IterType.BOTH

    def __init__(self, source: Iterable[T]):
        self.source = source
        source_iter_type = get_iter_type(source)

        # ソースの属性を継承し、クエリでタイプが強制された時はそのタイプを使う
        if self.iter_type == IterType.BOTH:
            self.run_iter_type = source_iter_type
        else:
            if self.iter_type == IterType.ASYNC:
                self.run_iter_type = self.iter_type

                if source_iter_type & IterType.ASYNC:
                    pass
                else:
                    # aiterのみ実行可能にする
                    self.source = QuerySyncToAsync(self.source)
            else:
                raise TypeError("can not convert sync iterator to any iteraotr.")

    def __iter__(self) -> Iterator[T]:
        if not (self.run_iter_type & IterType.NORMAL):
            raise NotImplementedError(f"{self.__class__} can't __iter__()")
        return self._impl_iter()

    def __aiter__(self) -> AsyncIterator[T]:
        if not (self.run_iter_type & IterType.ASYNC):
            raise NotImplementedError(f"{self.__class__} can't __aiter__()")
        return self._impl_aiter()

    def _impl_iter(self):
        return self.source.__iter__()

    def _impl_aiter(self):
        return self.source.__aiter__()

    to = finalizers.to


class QueryNormal(Query[T]):
    """同期イテレータを両対応するために使います"""

    iter_type = IterType.BOTH

    def __init__(self, source: Iterable[T]):
        # super().__init__(self)
        self.source = source
        self.run_iter_type = IterType.BOTH

        if not hasattr(source, "__iter__"):
            raise TypeError(f"{source} not has __iter__")

    def _impl_iter(self):
        return self.source.__iter__()

    @final
    async def _impl_aiter(self):
        for v in self._impl_iter():
            yield v


class QueryAsync(Query[T]):
    """非同期イテレータのみ対応のクエリ"""

    iter_type = IterType.ASYNC

    def __init__(self, source: Iterable[T]):
        self.source = source
        self.run_iter_type = IterType.ASYNC

        if not hasattr(source, "__aiter__"):
            raise TypeError(f"{source} not has __aiter__")

    def _impl_iter(self):
        raise NotImplementedError()

    def _impl_aiter(self):
        return self.source.__aiter__()


class QuerySeq(QueryNormal[T]):
    """リストなどをクエリ化します"""

    def _impl_iter(self):
        return self.source.__iter__()

    # def __reversed__(self):
    #     return self.source.__reversed__()


class QueryDict(QueryNormal[Tuple[K, V]]):
    """辞書などをクエリ化します"""

    if TYPE_CHECKING:

        def __init__(self, source: Mapping[K, V]):
            ...

    def _impl_iter(self):
        return self.source.items().__iter__()  # type: ignore

    # def __reversed__(self):
    #     return self.source.items().__reversed__()


class QuerySet(QueryNormal[T]):
    pass


async def sync_to_async_iterator(it):
    for x in it:
        yield x


class QuerySyncToAsync(Query[T]):
    """同期イテレータを非同期イテレータに変換します。
    もしくは、同期イテレータを取得できない場合、非同期イテレータの取得を試みます。"""

    iter_type = IterType.ASYNC

    def __init__(self, source: Iterable[T]):
        # super().__init__(self)
        self.source = source
        self.run_iter_type = self.iter_type

    def _impl_iter(self):
        raise NotImplementedError()

    def _impl_aiter(self):
        it = None
        try:
            it = iter(self.source)
        except Exception:
            pass

        if it:
            return sync_to_async_iterator(it)
        else:
            return self.source.__aiter__()



"""
Copyright (c) 2021 sasano8

Released under the MIT license.
see https://opensource.org/licenses/MIT

The universal_write_open function from:
https://github.com/EntilZha/PyFunctional/blob/master/functional/io.py
https://github.com/EntilZha/PyFunctional/blob/master/functional/pipeline.py
"""

import csv
import json
from functools import partial
from typing import Union

from typing_extensions import Literal

from ..io import universal_write_open


def to_file(
    iterable,
    path,
    delimiter=None,
    mode: Union[Literal["r", "w", "x", "a", "b", "t", "+"], str] = "wt",
    buffering=-1,
    encoding=None,
    errors=None,
    newline=None,
    compresslevel=9,
    format=None,
    check=-1,
    preset=None,
    filters=None,
    compression: Union[None, Literal["gz", "gzip", "lzma", "xz", "bz2"]] = None,
):
    """
    Saves the sequence to a file by executing str(self) which becomes str(self.to_list()). If
    delimiter is defined will instead execute self.make_string(delimiter)
    :param path: path to write file
    :param delimiter: if defined, will call make_string(delimiter) and save that to file.
    :param mode: file open mode
    :param buffering: passed to builtins.open
    :param encoding: passed to builtins.open
    :param errors: passed to builtins.open
    :param newline: passed to builtins.open
    :param compression: compression format
    :param compresslevel: passed to gzip.open
    :param format: passed to lzma.open
    :param check: passed to lzma.open
    :param preset: passed to lzma.open
    :param filters: passed to lzma.open
    """
    with universal_write_open(
        path,
        mode=mode,
        buffering=buffering,
        encoding=encoding,
        errors=errors,
        newline=newline,
        compression=compression,
        compresslevel=compresslevel,
        format=format,
        check=check,
        preset=preset,
        filters=filters,
    ) as output:
        if delimiter:
            output.write(iterable.make_string(delimiter))
        else:
            output.write(str(iterable))


def to_jsonl(
    iterable,
    path,
    mode: Union[Literal["r", "w", "x", "a", "b", "t", "+"], str] = "wb",
    compression: Union[None, Literal["gz", "gzip", "lzma", "xz", "bz2"]] = None,
    ensure_ascii=False,
):
    """
    Saves the sequence to a jsonl file. Each element is mapped using json.dumps then written
    with a newline separating each element.
    :param path: path to write file
    :param mode: mode to write in, defaults to 'w' to overwrite contents
    :param compression: compression format
    :param ensure_ascii: ensure_ascii
    """
    dumps = partial(json.dumps, ensure_ascii=ensure_ascii)
    with universal_write_open(path, mode=mode, compression=compression) as output:
        output.write((iterable.map(dumps).make_string("\n") + "\n").encode("utf-8"))


def to_json(
    iterable,
    path,
    root_array=True,
    mode: Union[Literal["r", "w", "x", "a", "b", "t", "+"], str] = "wt",
    compression: Union[None, Literal["gz", "gzip", "lzma", "xz", "bz2"]] = None,
    ensure_ascii=False,
):
    """
    Saves the sequence to a json file. If root_array is True, then the sequence will be written
    to json with an array at the root. If it is False, then the sequence will be converted from
    a sequence of (Key, Value) pairs to a dictionary so that the json root is a dictionary.
    :param path: path to write file
    :param root_array: write json root as an array or dictionary
    :param mode: file open mode
    """
    with universal_write_open(path, mode=mode, compression=compression) as output:
        if root_array:
            json.dump(list(iterable), output, ensure_ascii=ensure_ascii)
        else:
            json.dump(dict(iterable), output, ensure_ascii=ensure_ascii)


def to_csv(
    iterable,
    path,
    mode: Union[Literal["r", "w", "x", "a", "b", "t", "+"], str] = "wt",
    dialect="excel",
    compression: Union[None, Literal["gz", "gzip", "lzma", "xz", "bz2"]] = None,
    newline="",
    **fmtparams
):
    """
    Saves the sequence to a csv file. Each element should be an iterable which will be expanded
    to the elements of each row.
    :param path: path to write file
    :param mode: file open mode
    :param dialect: passed to csv.writer
    :param fmtparams: passed to csv.writer
    """

    if "b" in mode:
        newline = None

    with universal_write_open(
        path, mode=mode, compression=compression, newline=newline
    ) as output:
        csv_writer = csv.writer(output, dialect=dialect, **fmtparams)
        for row in iterable:
            csv_writer.writerow([str(element) for element in row])



import asyncio
from asyncio import iscoroutinefunction
from collections import deque
from functools import partial
from typing import AsyncIterable, TypeVar

from pnq import selectors
from pnq.concurrent import get_default_pool
from pnq.protocols import PExecutor
from pnq.selectors import starmap

from ..common import Listable
from .queries import _enumerate, chunk

T = TypeVar("T")


def _procceed(func, iterable):
    return [func(x) for x in iterable]


async def _procceed_async(func, iterable):
    return [await func(x) for x in iterable]


async def parallel(
    source: AsyncIterable[T], func, executor: PExecutor, *, unpack="", chunksize=1
):
    if executor is None:
        executor = get_default_pool()

    new_func = starmap(func, unpack)
    submit = executor.asubmit

    if executor.is_cpubound and chunksize != 1:
        runner = _procceed_async if asyncio.iscoroutine(func) else _procceed
        runner = partial(runner, new_func)

        tasks = [
            submit(runner, chuncked) async for chuncked in chunk(source, chunksize)
        ]
        for task in tasks:
            for x in await task:
                yield x

    else:
        tasks = [submit(new_func, x) async for x in source]
        for task in tasks:
            yield await task


async def dispatch(
    source: AsyncIterable[T],
    func,
    executor: PExecutor = None,
    *,
    unpack="",
    chunksize=1,
    on_complete=None
):
    if executor is None:
        executor = get_default_pool()

    if on_complete is None:
        on_complete = lambda x: x  # noqa

    new_func = starmap(func, unpack)
    submit = executor.asubmit

    if executor.is_cpubound and chunksize != 1:
        runner = _procceed_async if asyncio.iscoroutine(func) else _procceed
        runner = partial(runner, new_func)

        async for chuncked in chunk(source, chunksize):
            future = submit(runner, chuncked)
            future.add_done_callback(on_complete)
            await asyncio.sleep(0)

    else:
        async for x in source:
            future = submit(new_func, x)
            future.add_done_callback(on_complete)
            await asyncio.sleep(0)


def exec_request(func, *args, **kwargs):
    from ..requests import Response, StopWatch

    with StopWatch() as sw:
        err = None
        result = None
        try:
            result = func(*args, **kwargs)
        except Exception as e:
            err = e

    return Response(
        func,
        args,
        kwargs,
        err=err,
        res=result,
        start=sw.start,
        end=sw.end,
    )


async def exec_request_async(func, *args, **kwargs):
    from ..requests import Response, StopWatch

    with StopWatch() as sw:
        err = None
        result = None
        try:
            result = await func(*args, **kwargs)
        except Exception as e:
            err = e

    return Response(
        func,
        args,
        kwargs,
        err=err,
        res=result,
        start=sw.start,
        end=sw.end,
    )


def request(
    source: AsyncIterable[T],
    func,
    executor: PExecutor,
    *,
    unpack="",
    chunksize=1,
    retry: int = None,
    timeout: float = None
):
    if executor is None:
        executor = get_default_pool()

    if iscoroutinefunction(func):
        wrapped = partial(exec_request_async, func)
    else:
        wrapped = partial(exec_request, func)

    return parallel(source, wrapped, executor, unpack=unpack, chunksize=chunksize)


async def gather(
    source: AsyncIterable[T], parallel: int = 1, timeout=None, return_exceptions=True
):
    tasks = {}
    results = deque()

    def set_result(i, future):
        del tasks[i]
        results.append(future)

    async for i, x in _enumerate(
        Listable(source, selectors._select_future_with_schedule)
    ):
        x.add_done_callback(partial(set_result, i))
        tasks[i] = x
        while results:
            yield results.popleft().result()

    while tasks:
        while results:
            yield results.popleft().result()
        await asyncio.sleep(0.01)

    while results:
        yield results.popleft().result()


async def call_func(sem, x, timeout):
    async with sem:
        return await asyncio.wait_for(x, timeout)


async def gather_tagged(
    source: AsyncIterable[T], selector=None, parallel: int = 1, timeout=None
):
    if parallel > 1:
        tasks = []
        sem = asyncio.Semaphore(parallel)
        async for tag, x in Listable(source, selector):
            task = asyncio.create_task(call_func(sem, x, timeout))
            tasks.append((tag, task))
            await asyncio.sleep(0)

            if tasks[0][1].done():
                tag, task = tasks.pop(0)
                yield tag, task.result()

        while tasks:
            tag, task = tasks.pop(0)
            yield tag, await task

    else:
        async for tag, awaitable in Listable(source, selector):
            yield tag, await asyncio.wait_for(awaitable, timeout)



import asyncio
from concurrent.futures import Future as _Future


def _wrap_future(future: _Future) -> asyncio.Future:
    _future = asyncio.wrap_future(future)
    if not (future is _future):
        _future.__future__ = lambda: future  # type: ignore
    return _future


def as_future(target) -> _Future:
    if isinstance(target, _Future):
        return target
    return target.__future__()


def as_afuture(target) -> asyncio.Future:
    if isinstance(target, asyncio.Future):
        return target
    if isinstance(target, _Future):
        return _wrap_future(target)

    if hasattr(target, "__afuture__"):
        return target.__afuture__()
    else:
        return _wrap_future(target.__future__())



from decimal import Decimal, InvalidOperation
from typing import (
    Any,
    Callable,
    Coroutine,
    Generic,
    Iterable,
    NoReturn,
    Sequence,
    TypeVar,
    Union,
)

from pnq.exceptions import NoElementError, NotOneElementError

from .op import MAP_ASSIGN_OP, TH_ASSIGN_OP, TH_ROUND

T = TypeVar("T")


def to(self, finalizer):
    """ストリームをファイナライザによって処理します。

    Args:

    * self: 評価するシーケンス
    * finalizer: イテレータを受け取るクラス・関数

    Returns: ファイナライザが返す結果

    Usage:
    ```
    >>> pnq.query([1, 2, 3]).to(list)
    [1, 2]
    >>> pnq.query({1: "a", 2: "b"}).to(dict)
    {1: "a", 2: "b"}
    ```
    """
    return finalizer(self)


def map_if_not_none(it, selector):
    if selector is None:
        return it
    else:
        return map(selector, it)


def map_async_if_not_none(it, selector):
    if selector is None:
        return it
    else:
        return (selector(x) async for x in it)


class SyncFinalizer(Generic[T]):
    def __init__(self, source: Iterable[T]):
        self.source = source

    def __iter__(self):
        return self.source.__iter__()

    def to(self, finalizer):
        if False:  # is_async
            raise Exception()
        else:
            return finalizer(self)

    def lazy(self, finalizer):
        ...

    def len(self):
        if hasattr(self.source, "__len__"):
            return len(self.source)

        i = 0
        for i, v in enumerate(self, 1):
            ...

        return i

    def exists(self):
        for elm in self:
            return True

        return False

    def all(self, selector=lambda x: x):
        return all(map_if_not_none(self, selector))

    def any(self, selector=lambda x: x):
        return any(map_if_not_none(self, selector))

    def contains(self, value, selector=lambda x: x) -> bool:
        for val in map_if_not_none(self, selector):
            if val == value:
                return True

        return False

    def min(self, selector=lambda x: x, default=NoReturn):
        if default is NoReturn:
            return min(map_if_not_none(self, selector))
        else:
            return min(map_if_not_none(self, selector), default=default)

    def max(self, selector=lambda x: x, default=NoReturn):
        if default is NoReturn:
            return max(map_if_not_none(self, selector))
        else:
            return max(map_if_not_none(self, selector), default=default)

    def sum(self, selector=lambda x: x):
        return sum(map_if_not_none(self, selector))

    def average(
        self,
        selector=lambda x: x,
        exp: float = 0.00001,
        round: TH_ROUND = "ROUND_HALF_UP",
    ) -> Union[float, Decimal]:
        # import statistics
        # return statistics.mean(pmap(self, selector))  # type: ignore

        seed = Decimal("0")
        i = 0
        val = 0

        for i, val in enumerate(map_if_not_none(self, selector), 1):
            # val = selector(val)
            try:
                val = Decimal(str(val))  # type: ignore
            except InvalidOperation:
                raise TypeError(f"{val!r} is not a number")
            seed += val

        if i:
            result = seed / i
        else:
            result = seed

        result = result.quantize(Decimal(str(exp)), rounding=round)

        if isinstance(val, (int, float)):
            return float(result)
        else:
            return result

    def reduce(
        self,
        seed: T,
        op: Union[TH_ASSIGN_OP, Callable[[Any, Any], Any]] = "+=",
        selector=lambda x: x,
    ) -> T:
        if callable(op):
            binary_op = op
        else:
            binary_op = MAP_ASSIGN_OP[op]

        for val in map_if_not_none(self, selector):
            seed = binary_op(seed, val)

        return seed

    def concat(self, selector=lambda x: x, delimiter: str = ""):
        to_str = lambda x: "" if x is None else str(x)  # noqa
        return delimiter.join(to_str(x) for x in map_if_not_none(self, selector))

    def each(self, func=lambda x: x):
        for elm in self:
            func(elm)

    def each_unpack(self, func):
        for elm in self:
            func(**elm)

    async def each_async(self, func):
        for elm in self:
            await func(elm)

    async def each_async_unpack(self, func):
        for elm in self:
            await func(**elm)

    def one(self):
        it = iter(self)
        try:
            result = next(it)
        except StopIteration:
            raise NoElementError()

        try:
            next(it)
            raise NotOneElementError()
        except StopIteration:
            pass

        return result

    def first(self):
        it = iter(self)
        try:
            return next(it)
        except StopIteration:
            raise NoElementError()

    def last(self):
        if isinstance(self, Sequence):
            try:
                return self[-1]
            except IndexError:
                raise NoElementError()

        undefined = object()
        last = undefined
        for elm in self:
            last = elm

        if last is undefined:
            raise NoElementError()
        else:
            return last

    def one_or(self, default):
        try:
            # return one(self)
            return self.one()
        except NoElementError:
            return default

    def first_or(self, default):
        try:
            # return first(self)
            return self.first()
        except NoElementError:
            return default

    def last_or(self, default):
        try:
            # return last(self)
            return self.last()
        except NoElementError:
            return default

    def one_or_raise(self, exc: Union[str, Exception]):
        undefined = object()
        # result = one_or(self, undefined)
        result = self.one_or(undefined)
        if result is undefined:
            if isinstance(exc, str):
                raise Exception(exc)
            else:
                raise exc
        else:
            return result

    def first_or_raise(self, exc: Union[str, Exception]):
        undefined = object()
        # result = first_or(self, undefined)
        result = self.first_or(undefined)
        if result is undefined:
            if isinstance(exc, str):
                raise Exception(exc)
            else:
                raise exc
        else:
            return result

    def last_or_raise(self, exc: Union[str, Exception]):
        undefined = object()
        # result = last_or(self, undefined)
        result = self.last_or(undefined)
        if result is undefined:
            if isinstance(exc, str):
                raise Exception(exc)
            else:
                raise exc
        else:
            return result


class AsyncFinalizer(Generic[T]):
    def __init__(self, source: Iterable[T]):
        self.source = source

    def __aiter__(self):
        return self.source.__aiter__()

    async def to(self, finalizer):
        if finalizer is list:
            return [x async for x in self]
        elif finalizer is dict:
            return {k: v async for k, v in self}
        elif finalizer is set:
            return {x async for x in self}
        else:
            return await finalizer(self)

    async def lazy(self, finalizer):
        ...

    async def len(self) -> int:
        i = 0
        async for x in self:
            i += 1

        return i

    async def exists(self) -> bool:
        async for elm in self:
            return True

        return False

    async def all(self, selector=lambda x: x) -> bool:
        async for x in map_async_if_not_none(self, selector):
            if not x:
                return False
        return True

    async def any(self, selector=lambda x: x) -> bool:
        async for x in map_async_if_not_none(self, selector):
            if x:
                return True
        return False

    async def contains(self, value, selector=lambda x: x) -> bool:
        async for val in map_async_if_not_none(self, selector):
            if val == value:
                return True

        return False

    async def min(self, selector=lambda x: x, default=NoReturn) -> T:
        it = map_async_if_not_none(self, selector).__aiter__()

        try:
            _min = await it.__anext__()
        except StopAsyncIteration:
            if default is NoReturn:
                raise
            else:
                return default

        async for x in it:
            if x < _min:
                _min = x

        return _min

    async def max(self, selector=lambda x: x, default=NoReturn) -> T:
        it = map_async_if_not_none(self, selector).__aiter__()

        try:
            _max = await it.__anext__()
        except StopAsyncIteration:
            if default is NoReturn:
                raise
            else:
                return default

        async for x in it:
            if x > _max:
                _max = x

        return _max

    async def sum(self, selector=lambda x: x):
        it = map_async_if_not_none(self, selector).__aiter__()

        try:
            _sum = await it.__anext__()
        except StopIteration:
            return 0

        async for x in it:
            _sum += x

        return _sum

    async def average(
        self,
        selector=lambda x: x,
        exp: float = 0.00001,
        round: TH_ROUND = "ROUND_HALF_UP",
    ) -> Union[float, Decimal]:
        # import statistics
        # return statistics.mean(pmap(self, selector))  # type: ignore

        seed = Decimal("0")
        i = 0
        val = 0

        async for val in map_async_if_not_none(self, selector):
            i += 1

            try:
                val = Decimal(str(val))  # type: ignore
            except InvalidOperation:
                raise TypeError(f"{val!r} is not a number")
            seed += val

        if i:
            result = seed / i
        else:
            result = seed

        result = result.quantize(Decimal(str(exp)), rounding=round)

        if isinstance(val, (int, float)):
            return float(result)
        else:
            return result

    async def reduce(
        self,
        seed: T,
        op: Union[TH_ASSIGN_OP, Callable[[Any, Any], Any]] = "+=",
        selector=lambda x: x,
    ) -> T:
        if callable(op):
            binary_op = op
        else:
            binary_op = MAP_ASSIGN_OP[op]

        async for val in map_async_if_not_none(self, selector):
            seed = binary_op(seed, val)

        return seed

    async def concat(self, selector=lambda x: x, delimiter: str = "") -> str:
        to_str = lambda x: "" if x is None else str(x)  # noqa
        ait = (
            to_str(x) async for x in map_async_if_not_none(self, selector)
        ).__aiter__()

        # TODO: string.IOを使う
        try:
            current = await ait.__anext__()
        except StopAsyncIteration:
            return ""

        async for x in ait:
            current += delimiter + x

        return current

    async def each(self, func=lambda x: x) -> None:
        async for elm in self:
            func(elm)

    async def each_unpack(self, func) -> None:
        async for elm in self:
            func(**elm)

    async def each_async(self, func) -> None:
        async for elm in self:
            await func(elm)

    async def each_async_unpack(self, func) -> None:
        async for elm in self:
            await func(**elm)

    async def one(self) -> T:
        it = self.__aiter__()
        try:
            result = await it.__anext__()
        except StopAsyncIteration:
            raise NoElementError()

        try:
            await it.__anext__()
            raise NotOneElementError()
        except StopAsyncIteration:
            pass

        return result

    async def first(self) -> T:
        it = self.__aiter__()
        try:
            return await it.__anext__()
        except StopIteration:
            raise NoElementError()

    async def last(self):
        if isinstance(self, Sequence):
            try:
                return self[-1]
            except IndexError:
                raise NoElementError()

        undefined = object()
        last = undefined
        async for elm in self:
            last = elm

        if last is undefined:
            raise NoElementError()
        else:
            return last

    async def one_or(self, default):
        try:
            # return one(self)
            return await self.one()
        except NoElementError:
            return default

    async def first_or(self, default):
        try:
            # return first(self)
            return await self.first()
        except NoElementError:
            return default

    async def last_or(self, default):
        try:
            # return last(self)
            return await self.last()
        except NoElementError:
            return default

    async def one_or_raise(self, exc: Union[str, Exception]):
        undefined = object()
        # result = one_or(self, undefined)
        result = await self.one_or(undefined)
        if result is undefined:
            if isinstance(exc, str):
                raise Exception(exc)
            else:
                raise exc
        else:
            return result

    async def first_or_raise(self, exc: Union[str, Exception]):
        undefined = object()
        # result = first_or(self, undefined)
        result = await self.first_or(undefined)
        if result is undefined:
            if isinstance(exc, str):
                raise Exception(exc)
            else:
                raise exc
        else:
            return result

    async def last_or_raise(self, exc: Union[str, Exception]):
        undefined = object()
        # result = last_or(self, undefined)
        result = await self.last_or(undefined)
        if result is undefined:
            if isinstance(exc, str):
                raise Exception(exc)
            else:
                raise exc
        else:
            return result



import asyncio
import itertools
import time
from concurrent.futures import Future as ConcurrentFuture
from functools import partial
from typing import Iterable, Iterator

from typing_extensions import Literal

from .iterables import LazyListable


async def list_from_aiter(aiter):
    return [x async for x in aiter]


def map_iter(iterable, func, unpack: Literal["", "*", "**", "***"] = ""):
    if unpack == "":
        return (func(x) for x in iterable)
    elif unpack == "*":
        return (func(*x) for x in iterable)
    elif unpack == "**":
        return (func(**x) for x in iterable)
    elif unpack == "***":
        return (func(*x.args, **x.kwargs) for x in iterable)
    else:
        raise ValueError()


def map_iter_submit(
    executor, iterable, func, unpack: Literal["", "*", "**", "***"] = ""
) -> Iterator[ConcurrentFuture]:
    if unpack == "":
        return (executor.submit(func, x) for x in iterable)
    elif unpack == "*":
        return (executor.submit(func, *x) for x in iterable)
    elif unpack == "**":
        return (executor.submit(func, **x) for x in iterable)
    elif unpack == "***":
        return (executor.submit(func, *x.args, **x.kwargs) for x in iterable)
    else:
        raise ValueError()


def map_iter_asubmit(
    executor, iterable, func, unpack: Literal["", "*", "**", "***"] = ""
) -> Iterable[asyncio.Future]:
    if unpack == "":
        return (executor.asubmit(func, x) for x in iterable)
    elif unpack == "*":
        return (executor.asubmit(func, *x) for x in iterable)
    elif unpack == "**":
        return (executor.asubmit(func, **x) for x in iterable)
    elif unpack == "***":
        return (executor.asubmit(func, *x.args, **x.kwargs) for x in iterable)
    else:
        raise ValueError()


async def map_to_aiter(iterable, func, unpack: Literal["", "*", "**", "***"] = ""):
    if unpack == "":
        for x in iterable:
            yield await func(x)
    elif unpack == "*":
        for x in iterable:
            yield await func(*x)
    elif unpack == "**":
        for x in iterable:
            yield await func(**x)
    elif unpack == "***":
        for x in iterable:
            yield await func(*x.args, **x.kwargs)
    else:
        raise ValueError()


async def map_aiter(iterable, func, unpack: Literal["", "*", "**", "***"] = ""):
    if asyncio.iscoroutinefunction(func):
        if unpack == "":
            async for x in iterable:
                yield await func(x)
        elif unpack == "*":
            async for x in iterable:
                yield await func(*x)
        elif unpack == "**":
            async for x in iterable:
                yield await func(**x)
        elif unpack == "***":
            async for x in iterable:
                yield await func(*x.args, **x.kwargs)
        else:
            raise ValueError()
    else:
        if unpack == "":
            async for x in iterable:
                yield func(x)
        elif unpack == "*":
            async for x in iterable:
                yield func(*x)
        elif unpack == "**":
            async for x in iterable:
                yield func(**x)
        elif unpack == "***":
            async for x in iterable:
                yield func(*x.args, **x.kwargs)
        else:
            raise ValueError()


def process_chunk(func, chunk, unpack: Literal["", "*", "**"] = ""):
    return list(map_iter(chunk, func, unpack=unpack))


def process_chunk_async(
    func, chunk, unpack: Literal["", "*", "**"] = "", async_runner=None
):
    return async_runner(list_from_aiter(map_to_aiter(chunk, func, unpack=unpack)))


def take(iter, *args):
    return itertools.islice(iter, *args)


async def take_async(aiter, *args):
    r = range(*args)
    start = r.start
    stop = r.stop
    step = r.step

    current = 0
    it = aiter.__aiter__()

    try:
        while current < start:
            await it.__anext__()
            current += 1
    except StopAsyncIteration:
        return

    try:
        while current < stop:
            result = await it.__anext__()
            current += 1
            if (current % step) == 0:
                yield result

    except StopAsyncIteration:
        return


def chunk(iterable, chunksize: int):
    it = iter(iterable)
    while True:
        chunked = list(take(it, chunksize))
        if not chunked:
            return
        yield chunked


async def chunk_async(aiter, chunksize: int):
    it = aiter.__aiter__()
    while True:
        chunked = [x async for x in take_async(it, chunksize)]
        if not chunked:
            return
        yield chunked


def chain_chunks(iterable):
    for element in iterable:
        element.reverse()
        while element:
            yield element.pop()


def map_each_chunk(
    executor,
    iterable,
    func,
    unpack: Literal["", "*", "**"] = "",
    timeout=None,
    chunksize=1,
):
    if chunksize < 1:
        raise ValueError("chunksize must be >= 1.")

    if asyncio.iscoroutinefunction(func):
        processer = process_chunk_async
    else:
        processer = process_chunk

    results = map(
        executor,
        chunk(iterable, chunksize=chunksize),
        partial(processer, func, unpack=unpack),
        timeout=timeout,
    )
    return chain_chunks(results)


def map(
    executor,
    iterable,
    func,
    unpack: Literal["", "*", "**"] = "",
    timeout=None,
):
    return LazyListable(Map(executor, iterable, func, unpack, timeout))


class IterBase:
    def __iter__(self):
        return self._iter_impl()

    def __aiter__(self):
        return self._aiter_impl()

    @staticmethod
    def get_iter_type(iterable):
        if hasattr(iterable, "__aiter__"):
            if hasattr(iterable, "__iter__"):
                return 3
            else:
                return 2
        else:
            if hasattr(iterable, "__iter__"):
                return 1
            else:
                return 0

    @classmethod
    def _get_is_aiter(cls, iterable):
        type = cls.get_iter_type(iterable)
        if type == 2:
            return True
        elif type == 1:
            return False
        elif type == 3:
            raise TypeError(
                f"Could not identify {iterable} as '__iter__' or '__aiter__'"
            )
        else:
            raise TypeError(f"{iterable} has no '__iter__' or '__aiter__'")

    def _iter_impl(self):
        raise NotImplementedError()

    def _aiter_impl(self):
        raise NotImplementedError()


class Chunk(IterBase):
    def __init__(
        self,
        iterable,
        chunksize=1,
    ):
        self.source = iterable
        self.chunksize = chunksize

    def _iter_impl(self):
        it = iter(self.source)
        while True:
            chunked = list(itertools.islice(it, self.chunksize))
            if not chunked:
                return
            yield chunked

    async def _aiter_impl(self):
        it = self.source.__aiter__()
        while True:
            chunked = [x async for x in take_async(it, self.chunksize)]
            if not chunked:
                return
            yield chunked


# TODO: 塊ごとに１ワーカーで処理するのか、要素ごとに１ワーカーで処理するのか使い分けれるといい
class Map(IterBase):
    def __init__(
        self,
        executor,
        iterable,
        func,
        unpack: Literal["", "*", "**"] = "",
        timeout=None,
        chunksize=None,
    ):
        if chunksize is None:
            chunksize = executor.max_workers

        self._get_is_aiter(iterable)

        self.executor = executor
        self.source = iterable
        self.func = func
        self.unpack = unpack
        self.timeout = timeout

    def __iter__(self):
        if getattr(self.executor, "is_async_only", False):
            raise TypeError(
                f"{self.executor} are only allowed async. use async for or await."
            )
        return self._iter_impl()

    def _iter_impl(self):
        return self._get_iter(
            self.executor,
            self.source,
            self.func,
            self.unpack,
            self.timeout,
        )

    def _aiter_impl(self):
        return self._get_aiter(
            self.executor,
            self.source,
            self.func,
            self.unpack,
            self.timeout,
        )

    @classmethod
    def _get_iter(cls, executor, it, func, unpack, timeout, chunksize=1):
        if timeout is not None:
            end_time = timeout + time.monotonic()

        def result_iterator(chunk: list):
            try:
                # reverse to keep finishing order
                fs.reverse()
                while fs:
                    # Careful not to keep a reference to the popped future
                    if timeout is None:
                        yield fs.pop().result()
                    else:
                        yield fs.pop().result(end_time - time.monotonic())
            finally:
                for future in fs:
                    future.cancel()

        for chunked in Chunk(it, chunksize):
            fs = list(map_iter_submit(executor, chunked, func, unpack=unpack))
            for x in result_iterator(fs):
                yield x

    @classmethod
    async def _get_aiter(cls, executor, it, func, unpack, timeout, chunksize=1):
        if timeout is not None:
            end_time = timeout + time.monotonic()

        async def result_iterator(chunk: list):
            try:
                # reverse to keep finishing order
                fs.reverse()
                while fs:
                    # Careful not to keep a reference to the popped future
                    if timeout is None:
                        yield await fs.pop()
                    else:
                        yield await asyncio.wait_for(
                            fs.pop(), timeout=end_time - time.monotonic()
                        )
            finally:
                for future in fs:
                    future.cancel()

        is_async = cls._get_is_aiter(it)

        if is_async:
            async for chunked in Chunk(it, chunksize):
                fs = list(map_iter_asubmit(executor, chunked, func, unpack=unpack))
                async for x in result_iterator(fs):
                    yield x

        else:
            for chunked in Chunk(it, chunksize):
                fs = list(map_iter_asubmit(executor, chunked, func, unpack=unpack))
                async for x in result_iterator(fs):
                    yield x



import asyncio
from decimal import Decimal, InvalidOperation
from typing import Any, Callable, Iterable, NoReturn, Sequence, TypeVar, Union

from pnq._itertools.common import Listable, name_as
from pnq._itertools.op import MAP_ASSIGN_OP, TH_ASSIGN_OP, TH_ROUND
from pnq.exceptions import NoElementError, NotOneElementError
from pnq.selectors import starmap

from . import queries

T = TypeVar("T")


def to(source, finalizer):
    return finalizer(source)


@name_as("len")
def _len(source: Iterable[T]) -> int:
    count = 0
    for x in Listable(source, None):
        count += 1

    return count


def exists(source: Iterable[T], predicate=None) -> bool:
    for x in queries._filter(source, predicate):
        return True

    return False


def empty(source: Iterable[T], predicate=None) -> bool:
    for x in queries._filter(source, predicate):
        return False

    return True


@name_as("all")
def _all(source: Iterable[T], selector=None) -> bool:
    for x in Listable(source, selector):
        if not x:
            return False

    return True


@name_as("any")
def _any(source: Iterable[T], selector=None) -> bool:
    for x in Listable(source, selector):
        if x:
            return True

    return False


# containsより意思的
def find(source: Iterable[T], value, selector=None) -> bool:
    for val in Listable(source, selector):
        if val == value:
            return True

    return False


def contains(source: Iterable[T], value, selector=None) -> bool:
    for val in Listable(source, selector):
        if val == value:
            return True

    return False


@name_as("sum")
def _sum(source: Iterable[T], selector=None):
    current = 0
    for val in Listable(source, selector):
        current += val

    return current


@name_as("min")
def _min(source: Iterable[T], key_selector=None, default=NoReturn):
    if default is NoReturn:
        return min(Listable(source), key=key_selector)
    else:
        return min(Listable(source), key=key_selector, default=default)


@name_as("max")
def _max(source: Iterable[T], key_selector=None, default=NoReturn):
    if default is NoReturn:
        return max(Listable(source), key=key_selector)
    else:
        return max(Listable(source), key=key_selector, default=default)


def average(
    source: Iterable[T],
    selector=lambda x: x,
    exp: float = 0.00001,
    round: TH_ROUND = "ROUND_HALF_UP",
) -> Union[float, Decimal]:
    # import statistics
    # return statistics.mean(pmap(self, selector))  # type: ignore

    seed = Decimal("0")
    i = 0
    val = 0

    for val in Listable(source, selector):
        i += 1
        # val = selector(val)
        try:
            val = Decimal(str(val))  # type: ignore
        except InvalidOperation:
            raise TypeError(f"{val!r} is not a number")
        seed += val

    if i:
        result = seed / i
    else:
        result = seed

    result = result.quantize(Decimal(str(exp)), rounding=round)

    if isinstance(val, (int, float)):
        return float(result)
    else:
        return result


def reduce(
    source: Iterable[T],
    seed: Union[T, None],
    op: Union[TH_ASSIGN_OP, Callable[[Any, Any], Any]] = "+=",
    selector=lambda x: x,
) -> T:
    if isinstance(op, str):
        binary_op = MAP_ASSIGN_OP[op]
    else:
        binary_op = op

    it = Listable(source, selector).__iter__()

    if seed is None:
        for val in it:
            seed = val
            break
        else:
            raise TypeError("empty sequence with no seed.")

    for val in Listable(source, selector):
        seed = binary_op(seed, val)

    return seed  # type: ignore


def accumulate(
    source: Iterable[T],
    seed: Union[T, None],
    op: Union[TH_ASSIGN_OP, Callable[[Any, Any], Any]] = "+=",
    selector=lambda x: x,
):
    if isinstance(op, str):
        binary_op = MAP_ASSIGN_OP[op]
    else:
        binary_op = op

    it = Listable(source, selector).__iter__()

    results = []

    if seed is None:
        for val in it:
            seed = val
            results.append(seed)
            break
        else:
            return results
    else:
        results.append(seed)

    for val in it:
        seed = binary_op(seed, val)
        results.append(seed)

    return results


def concat(source: Iterable[T], selector=None, delimiter: str = ""):
    to_str = lambda x: "" if x is None else str(x)  # noqa
    return delimiter.join(Listable(Listable(source, selector), to_str))


def each(source: Iterable[T], func=lambda x: x, unpack=""):
    if asyncio.iscoroutinefunction(func):
        func = starmap(func, unpack)
        for elm in source:
            func(elm)
    else:
        func = starmap(func, unpack)
        for elm in source:
            func(elm)


def one(source: Iterable[T]):
    it = source.__iter__()
    try:
        result = it.__next__()
    except StopIteration:
        raise NoElementError()

    try:
        it.__next__()
        raise NotOneElementError()
    except StopIteration:
        pass

    return result


def first(source: Iterable[T]):
    it = source.__iter__()
    try:
        return it.__next__()
    except StopIteration:
        raise NoElementError()


def last(source: Iterable[T]):
    if isinstance(source, Sequence):
        try:
            return source[-1]
        except IndexError:
            raise NoElementError()

    undefined = object()
    last = undefined
    for elm in source:
        last = elm

    if last is undefined:
        raise NoElementError()
    else:
        return last


def one_or(source: Iterable[T], default):
    try:
        return one(source)
    except NoElementError:
        return default


def first_or(source: Iterable[T], default):
    try:
        return first(source)
    except NoElementError:
        return default


def last_or(source: Iterable[T], default):
    try:
        return last(source)
    except NoElementError:
        return default


def one_or_raise(source: Iterable[T], exc: Union[str, Exception]):
    undefined = object()
    result = one_or(source, undefined)
    if result is undefined:
        if isinstance(exc, str):
            raise Exception(exc)
        else:
            raise exc
    else:
        return result


def first_or_raise(source: Iterable[T], exc: Union[str, Exception]):
    undefined = object()
    result = first_or(source, undefined)
    if result is undefined:
        if isinstance(exc, str):
            raise Exception(exc)
        else:
            raise exc
    else:
        return result


def last_or_raise(source: Iterable[T], exc: Union[str, Exception]):
    undefined = object()
    result = last_or(source, undefined)
    if result is undefined:
        if isinstance(exc, str):
            raise Exception(exc)
        else:
            raise exc
    else:
        return result


def to_dict(key_values):
    if hasattr(key_values, "__iter__"):
        result = {}
        for k, v in key_values:
            result[k] = v
        return result
    elif hasattr(key_values, "__iter__"):
        return dict(key_values)
    else:
        raise TypeError(f"{key_values} no has __iter__ or __aiter__")



from pnq.protocols import Executor, PExecutable, PExecutor  # noqa

# import asyncio
# import concurrent
# from typing import Dict, Iterable

# from typing_extensions import Protocol, runtime_checkable  # from python3.8

# """
# concurrent.futures.ProcessPoolExecutorはmultiprocessing.Poolのラッパーです。
# https://stackoverflow.com/questions/38311431/concurrent-futures-processpoolexecutor-vs-multiprocessing-pool-pool
# """


# @runtime_checkable
# class Executor(Protocol):
#     def submit(self, *args, **kwargs) -> concurrent.futures.Future:
#         ...


# class PExecutor(Protocol):
#     def submit(self, func, *args, **kwargs) -> concurrent.futures.Future:
#         ...

#     def asubmit(self, func, *args, **kwargs) -> asyncio.Future:
#         ...

#     @property
#     def running_task_count(self) -> int:
#         ...

#     @property
#     def is_full(self) -> bool:
#         ...

#     @property
#     def is_closed(self) -> bool:
#         ...

#     @property
#     def max_workers(self) -> int:
#         ...

#     @property
#     def is_cpubound(self) -> bool:
#         ...


# class PExecutable(Protocol):
#     def __executor__(self) -> PExecutor:
#         ...



import asyncio
from contextlib import AsyncExitStack, ExitStack
from functools import partial
from typing import TYPE_CHECKING, Iterable

from .protocols import PArguments


class Arguments(PArguments):
    """引数を保持するためのクラスです。

    Usage:
    ```
    >>> def show(a1, a2, kw1, kw2):
    >>>   print(a1, a2, kw1, kw2)
    >>> a = pnq.args(1, 2, kw1=3, kw2=4)
    >>> pnq.query([a]).each(show, unpack="***")
    1, 2, 3, 4
    ```
    """

    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

    @classmethod
    def from_obj(cls, args, kwargs):
        obj = cls()
        obj.args = args
        obj.kwargs = kwargs
        return obj

    def __iter__(self):
        return iter(self.args)

    def __getitem__(self, key):
        return self.kwargs[key]

    def keys(self):
        return self.kwargs.keys()

    def values(self):
        return self.kwargs.values()

    def items(self):
        return self.kwargs.items()

    def partial(self, func):
        return partial(func, *self.args, **self.kwargs)


class PnqExitStack:
    def __init__(self, contexts: Iterable):
        self.source = contexts
        self._init = False

    @classmethod
    def from_args(cls, *contexts):
        return cls(contexts)

    def __iter__(self):
        return iter(self.source)

    def start(self):
        if self._init:
            raise RuntimeError("ExitStack is already initialized")

        contexts = []
        with ExitStack() as stack:
            for item in self.source:
                stack.enter_context(item)
                contexts.append(item)

            self.close = stack.pop_all().close  # type: ignore

        self._init = True
        return tuple(contexts)

    async def astart(self):
        if self._init:
            raise RuntimeError("ExitStack is already initialized")

        contexts = []
        async with AsyncExitStack() as stack:
            for item in self.source:
                await stack.enter_async_context(item)
                contexts.append(item)

            self.aclose = stack.pop_all().aclose  # type: ignore

        self._init = True
        return tuple(contexts)

    if TYPE_CHECKING:

        def close(self):
            ...

        async def aclose(self):
            ...

    def __enter__(self):
        return self.start()

    def __exit__(self, *args, **kwargs):
        self.close()

    async def __aenter__(self):
        return await self.astart()

    async def __aexit__(self, *args, **kwargs):
        await self.aclose()


exitstack = PnqExitStack.from_args


class PnqQueue:
    def __init__(self) -> None:
        from collections import deque

        self._items = deque()
        self._running = True

    def stop(self):
        self._running = False

    def kill(self):
        self._running = False
        self._items.clear()

    def put(self, item):
        if not self._running:
            raise RuntimeError("Queue is stopped")
        self._items.append(item)

    def try_pop(self):
        if not self._running:
            raise RuntimeError("Queue is stopped")
        if not self._items:
            return False, None
        return True, self._items.popleft()

    async def __aiter__(self):
        if not self._running:
            raise RuntimeError("Queue is stopped")

        sleep = asyncio.sleep
        queue = self._items

        while self._running:
            if not queue:
                await sleep(0)
                continue
            yield queue.popleft()

        while queue:
            yield queue.popleft()



import json
from contextlib import _GeneratorContextManager
from functools import wraps
from typing import Any, List, cast

from pnq.queries import QueryBase as Query
from pnq.queries import query


class PnqContextManager(_GeneratorContextManager):
    def __iter__(self):
        with self as iter:
            for x in iter:
                yield x


def contextmanager(func):
    @wraps(func)
    def helper(*args, **kwds):
        return PnqContextManager(func, args, kwds)  # type: ignore

    return helper


@contextmanager
def from_csv(path, delimiter=","):
    import csv

    with open(path) as f:
        reader = csv.reader(f, delimiter=delimiter)
        yield cast(Query[List[str]], query(reader))


@contextmanager
def from_jsonl(path, deserializer=json.loads):
    with open(path) as f:
        strip_lines = (line.strip() for line in f.readlines())
        ignore_empty_lines = (deserializer(x) for x in strip_lines if x)
        yield cast(Query[Any], query(ignore_empty_lines))



import asyncio
from decimal import Decimal, InvalidOperation
from typing import Any, AsyncIterable, Callable, NoReturn, Sequence, TypeVar, Union

from pnq._itertools.common import Listable, name_as
from pnq._itertools.op import MAP_ASSIGN_OP, TH_ASSIGN_OP, TH_ROUND
from pnq.exceptions import NoElementError, NotOneElementError
from pnq.selectors import starmap

from . import queries

T = TypeVar("T")


async def to(source, finalizer):
    return await finalizer(source)


@name_as("len")
async def _len(source: AsyncIterable[T]) -> int:
    count = 0
    async for x in Listable(source, None):
        count += 1

    return count


async def exists(source: AsyncIterable[T], predicate=None) -> bool:
    async for x in queries._filter(source, predicate):
        return True

    return False


async def empty(source: AsyncIterable[T], predicate=None) -> bool:
    async for x in queries._filter(source, predicate):
        return False

    return True


@name_as("all")
async def _all(source: AsyncIterable[T], selector=None) -> bool:
    async for x in Listable(source, selector):
        if not x:
            return False

    return True


@name_as("any")
async def _any(source: AsyncIterable[T], selector=None) -> bool:
    async for x in Listable(source, selector):
        if x:
            return True

    return False


# containsより意思的
async def find(source: AsyncIterable[T], value, selector=None) -> bool:
    async for val in Listable(source, selector):
        if val == value:
            return True

    return False


async def contains(source: AsyncIterable[T], value, selector=None) -> bool:
    async for val in Listable(source, selector):
        if val == value:
            return True

    return False


@name_as("sum")
async def _sum(source: AsyncIterable[T], selector=None):
    current = 0
    async for val in Listable(source, selector):
        current += val

    return current


@name_as("min")
async def _min(source: AsyncIterable[T], key_selector=None, default=NoReturn):
    if default is NoReturn:
        return min(await Listable(source), key=key_selector)
    else:
        return min(await Listable(source), key=key_selector, default=default)


@name_as("max")
async def _max(source: AsyncIterable[T], key_selector=None, default=NoReturn):
    if default is NoReturn:
        return max(await Listable(source), key=key_selector)
    else:
        return max(await Listable(source), key=key_selector, default=default)


async def average(
    source: AsyncIterable[T],
    selector=lambda x: x,
    exp: float = 0.00001,
    round: TH_ROUND = "ROUND_HALF_UP",
) -> Union[float, Decimal]:
    # import statistics
    # return statistics.mean(pmap(self, selector))  # type: ignore

    seed = Decimal("0")
    i = 0
    val = 0

    async for val in Listable(source, selector):
        i += 1
        # val = selector(val)
        try:
            val = Decimal(str(val))  # type: ignore
        except InvalidOperation:
            raise TypeError(f"{val!r} is not a number")
        seed += val

    if i:
        result = seed / i
    else:
        result = seed

    result = result.quantize(Decimal(str(exp)), rounding=round)

    if isinstance(val, (int, float)):
        return float(result)
    else:
        return result


async def reduce(
    source: AsyncIterable[T],
    seed: Union[T, None],
    op: Union[TH_ASSIGN_OP, Callable[[Any, Any], Any]] = "+=",
    selector=lambda x: x,
) -> T:
    if isinstance(op, str):
        binary_op = MAP_ASSIGN_OP[op]
    else:
        binary_op = op

    it = Listable(source, selector).__aiter__()

    if seed is None:
        async for val in it:
            seed = val
            break
        else:
            raise TypeError("empty sequence with no seed.")

    async for val in Listable(source, selector):
        seed = binary_op(seed, val)

    return seed  # type: ignore


async def accumulate(
    source: AsyncIterable[T],
    seed: Union[T, None],
    op: Union[TH_ASSIGN_OP, Callable[[Any, Any], Any]] = "+=",
    selector=lambda x: x,
):
    if isinstance(op, str):
        binary_op = MAP_ASSIGN_OP[op]
    else:
        binary_op = op

    it = Listable(source, selector).__aiter__()

    results = []

    if seed is None:
        async for val in it:
            seed = val
            results.append(seed)
            break
        else:
            return results
    else:
        results.append(seed)

    async for val in it:
        seed = binary_op(seed, val)
        results.append(seed)

    return results


async def concat(source: AsyncIterable[T], selector=None, delimiter: str = ""):
    to_str = lambda x: "" if x is None else str(x)  # noqa
    return delimiter.join(await Listable(Listable(source, selector), to_str))


async def each(source: AsyncIterable[T], func=lambda x: x, unpack=""):
    if asyncio.iscoroutinefunction(func):
        func = starmap(func, unpack)
        async for elm in source:
            await func(elm)
    else:
        func = starmap(func, unpack)
        async for elm in source:
            func(elm)


async def one(source: AsyncIterable[T]):
    it = source.__aiter__()
    try:
        result = await it.__anext__()
    except StopAsyncIteration:
        raise NoElementError()

    try:
        await it.__anext__()
        raise NotOneElementError()
    except StopAsyncIteration:
        pass

    return result


async def first(source: AsyncIterable[T]):
    it = source.__aiter__()
    try:
        return await it.__anext__()
    except StopAsyncIteration:
        raise NoElementError()


async def last(source: AsyncIterable[T]):
    if isinstance(source, Sequence):
        try:
            return source[-1]
        except IndexError:
            raise NoElementError()

    undefined = object()
    last = undefined
    async for elm in source:
        last = elm

    if last is undefined:
        raise NoElementError()
    else:
        return last


async def one_or(source: AsyncIterable[T], default):
    try:
        return await one(source)
    except NoElementError:
        return default


async def first_or(source: AsyncIterable[T], default):
    try:
        return await first(source)
    except NoElementError:
        return default


async def last_or(source: AsyncIterable[T], default):
    try:
        return await last(source)
    except NoElementError:
        return default


async def one_or_raise(source: AsyncIterable[T], exc: Union[str, Exception]):
    undefined = object()
    result = await one_or(source, undefined)
    if result is undefined:
        if isinstance(exc, str):
            raise Exception(exc)
        else:
            raise exc
    else:
        return result


async def first_or_raise(source: AsyncIterable[T], exc: Union[str, Exception]):
    undefined = object()
    result = await first_or(source, undefined)
    if result is undefined:
        if isinstance(exc, str):
            raise Exception(exc)
        else:
            raise exc
    else:
        return result


async def last_or_raise(source: AsyncIterable[T], exc: Union[str, Exception]):
    undefined = object()
    result = await last_or(source, undefined)
    if result is undefined:
        if isinstance(exc, str):
            raise Exception(exc)
        else:
            raise exc
    else:
        return result


async def to_dict(key_values):
    if hasattr(key_values, "__aiter__"):
        result = {}
        async for k, v in key_values:
            result[k] = v
        return result
    elif hasattr(key_values, "__iter__"):
        return dict(key_values)
    else:
        raise TypeError(f"{key_values} no has __iter__ or __aiter__")



from .executors import AsyncPoolExecutor as AsyncPool
from .executors import DummyPoolExecutor as DummyPool
from .executors import ProcessPoolExecutor as ProcessPool
from .executors import ThreadPoolExecutor as ThreadPool
from .executors import get_default_pool, get_executor
from .protocols import PExecutor
from .tools import map



def name_as(name):
    def wrapper(func):
        func.__name__ = name
        return func

    return wrapper


class Listable:
    def __init__(self, source, selector=None) -> None:
        self.source = source
        self.selector = selector

    def __iter__(self):
        selector = self.selector
        it = self.source.__iter__()

        if selector is None:
            return it
        else:
            return (selector(x) for x in it)

    def __aiter__(self):
        selector = self.selector
        ait = self.source.__aiter__()

        if selector is None:
            return ait
        else:
            return (selector(x) async for x in ait)

    def __await__(self):
        return self._result_async().__await__()

    def result(self, timeout=None):
        return list(self)

    async def _result_async(self):
        return [x async for x in self]



from typing import (
    TYPE_CHECKING,
    Any,
    AsyncGenerator,
    AsyncIterable,
    AsyncIterator,
    Awaitable,
    Callable,
    Dict,
    FrozenSet,
    Generator,
    Generic,
    Iterable,
    Iterator,
    List,
    Mapping,
    NoReturn,
    Sequence,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
    no_type_check,
    overload,
)

try:
    from typing import Literal
except:
    from typing_extensions import Literal

from pnq.exceptions import NotFoundError

from ._itertools import builder, core, queryables
from ._itertools.op import TH_ASSIGN_OP
from ._itertools.queryables import AsyncFinalizer, Finalizer
from ._itertools.requests import Response

T = TypeVar("T")
K = TypeVar("K")
V = TypeVar("V")
K2 = TypeVar("K2")
V2 = TypeVar("V2")
R = TypeVar("R")


__all__ = ["Query", "PairQuery", "query"]


{% for query in queries %}

class {{query.CLS}}:
    if TYPE_CHECKING:
        def __iter__(self) -> Iterator[{{query.T}}]:
            ...

        def __aiter__(self) -> AsyncIterator[{{query.T}}]:
            ...

    def as_iter(self) -> "Finalizer[{{query.T}}]":
        # return finalizers.SyncFinalizer(self)
        return Finalizer(self)

    def as_aiter(self) -> "AsyncFinalizer[{{query.T}}]":
        return AsyncFinalizer(self)

    to_file = Finalizer.to_file
    to_csv = Finalizer.to_csv
    to_json = Finalizer.to_json
    to_jsonl = Finalizer.to_jsonl

    @property
    def _(self) -> "AsyncFinalizer[{{query.T}}]":
        return AsyncFinalizer(self)

    async def __aresult__(self):
        return await PnqList.from_aiter(self)

    {% if query.is_pair %}

    def result(self, timeout=None) -> "PnqListPair[{{query.K}}, {{query.V}}]":
        return PnqList(self)

    save = result

    def __await__(self) -> Generator[Any, Any, "PnqListPair[{{query.K}}, {{query.V}}]"]:
        return self.__aresult__().__await__()

    {% else %}

    def result(self, timeout=None) -> "PnqList[{{query.T}}]":
        return PnqList(self)

    save = result

    def __await__(self) -> Generator[Any, Any, "PnqList[{{query.T}}]"]:
        return self.__aresult__().__await__()

    {% endif %}

    def len(self) -> int:
        return Finalizer.len(self)

    def exists(self, predicate=None) -> bool:
        return Finalizer.exists(self, predicate)

    def all(self, selector: Callable[[{{query.T}}], Any]=lambda x: x) -> bool:
        return Finalizer.all(self, selector)

    def any(self, selector: Callable[[{{query.T}}], Any]=lambda x: x) -> bool:
        return Finalizer.any(self, selector)

    def contains(self, value, selector: Callable[[{{query.T}}], Any]=lambda x: x) -> bool:
        return Finalizer.contains(self, value, selector)

    @overload
    def min(self, *, default=NoReturn) -> Union[{{query.T}}, NoReturn]: ...

    @overload
    def min(self, key_selector: Callable[[{{query.T}}], R]=lambda x: x, default=NoReturn) -> R: ...
    def min(self, key_selector: Callable[[{{query.T}}], R]=lambda x: x, default=NoReturn) -> R:
        return Finalizer.min(self, key_selector, default)

    @overload
    def max(self, *, default=NoReturn) -> Union[{{query.T}}, NoReturn]: ...
    @overload
    def max(self, key_selector: Callable[[{{query.T}}], R]=lambda x: x, default=NoReturn) -> R: ...
    def max(self, key_selector: Callable[[{{query.T}}], R]=lambda x: x, default=NoReturn) -> R:
        return Finalizer.max(self, key_selector, default)

    @overload
    def sum(self) -> {{query.T}}: ...
    @overload
    def sum(self, selector: Callable[[{{query.T}}], R]=lambda x: x) -> R: ...
    def sum(self, selector: Callable[[{{query.T}}], R]=lambda x: x) -> R:
        return Finalizer.sum(self, selector)

    @overload
    def average(self) -> {{query.T}}: ...
    @overload
    def average(self, selector: Callable[[{{query.T}}], R]=lambda x: x) -> R: ...
    def average(self, selector: Callable[[{{query.T}}], R]=lambda x: x) -> R:
        return Finalizer.average(self, selector)

    def reduce(self, seed: T, op: Union[TH_ASSIGN_OP, Callable[[Any, Any], Any]] = "+=", selector=lambda x: x) -> T:
        return Finalizer.reduce(self, seed, op, selector)

    def concat(self, selector=lambda x: x, delimiter: str = "") -> str:
        return Finalizer.concat(self, selector, delimiter)

    def test(self, expect):
        """同期イテレータと非同期イテレータを実行し結果を比較し、
        それぞれの比較結果(bool)をタプルとして返す。
        実行不可な場合はNoneを返す。
        """

        # TODO: asyncを通るようにパスを切り替えるクエリを挟む必要がある

        import asyncio

        async def to_list():
            return [x async for x in self]

        sync_result = None
        async_result = None

        try:
            sync_result = [x for x in self]
        except Exception as e:
            pass

        try:
            async_result = asyncio.run(to_list())
        except Exception as e:
            pass

        return_sync = None
        return_async = None

        if sync_result is not None:
            return_sync = (sync_result == expect)

        if async_result is not None:
            return_async = (async_result == expect)

        return return_sync, return_async

    def gather(self):
        return queryables.Gather(self)

    def request(
        self,
        func,
        executor=None,
        *,
        unpack="",
        chunksize=1,
        retry: int = None,
        timeout: float = None,
    ) -> "Query[Response]":
        return queryables.Request(self, func, executor, unpack=unpack, chunksize=chunksize, retry=retry, timeout=timeout)

    @overload
    def parallel(self, func: Callable[..., Awaitable[R]], executor=None, *, unpack="", chunksize=1) -> "Query[R]":
        ...

    @overload
    def parallel(self, func: Callable[..., R], executor=None, *, unpack="", chunksize=1) -> "Query[R]":
        ...

    def parallel(self, func, executor=None, *, unpack="", chunksize=1) -> "Query[R]":
        return queryables.Parallel(self, func, executor, unpack=unpack, chunksize=chunksize)

    def dispatch(self, func, executor: "PExecutor" = None, *, unpack="", chunksize=1, on_complete=None):
        return Finalizer.dispatch(self, func, executor, unpack=unpack, chunksize=chunksize, on_complete=on_complete)

    {% if query.is_pair %}

    @overload
    def to(self: Iterable[Tuple[K, V]], func: Type[Mapping[K, V]]) -> Mapping[K, V]:
        ...

    @overload
    def to(self: Iterable[Tuple[K, V]], func: Callable[[Iterable[Tuple[K, V]]], R]) -> R:
        ...

    {% endif %}

    @overload
    def to(self: Iterable[T], func: Type[Iterable[T]]) -> Iterable[T]:
        ...

    @overload
    def to(self: Iterable[T], func: Callable[[Iterable[T]], R]) -> R:
        ...

    def to(self, func: Callable[[Iterable[T]], R]) -> R:
        return Finalizer.to(self, func)

    def lazy(self, func, *args, **kwargs):
        return Finalizer.lazy(self, func, *args, **kwargs)

    def each(self, func: Callable = lambda x: x, unpack=""):
        return Finalizer.each(self, func, unpack)

    def one(self) -> {{query.T}}:
        return Finalizer.one(self)

    def one_or(self, default: R) -> Union[{{query.T}}, R]:
        return Finalizer.one_or(self, default)

    def one_or_raise(self, exc: Union[str, Exception]) -> {{query.T}}:
        return Finalizer.one_or_raise(self, exc)

    def first(self) -> {{query.T}}:
        return Finalizer.first(self)

    def first_or(self, default: R) -> Union[{{query.T}}, R]:
        return Finalizer.first_or(self, default)

    def first_or_raise(self, exc: Union[str, Exception]) -> {{query.T}}:
        return Finalizer.first_or_raise(self, exc)

    def last(self) -> {{query.T}}:
        return Finalizer.last(self)

    def last_or(self, default: R) -> Union[{{query.T}}, R]:
        return Finalizer.last_or(self, default)

    def last_or_raise(self, exc: Union[str, Exception]) -> {{query.T}}:
        return Finalizer.last_or_raise(self, exc)

    @overload
    def cast(self, type: Type[Tuple[K2, V2]]) -> "{{pair.SELF__}}[K2, V2]":
        pass

    @overload
    def cast(self, type: Type[R]) -> "Query[R]":
        pass

    def cast(self, type: Type[R]) -> "Query[R]":
        return self

    def enumerate(self, start: int = 0, step: int = 1) -> "{{pair.SELF__}}[int, {{query.T}}]":
        return queryables.Enumerate(self, start, step)

    @overload
    def map(self, selector: Callable[[{{query.T}}], Tuple[K2, V2]], unpack="") -> "{{pair.SELF__}}[K2, V2]":
        pass

    @overload
    def map(self, selector: Callable[[{{query.T}}], R], unpack="") -> "{{sequence.SELF__}}[R]":
        pass

    @overload
    def map(self, selector: Callable[..., R], unpack="*") -> "{{sequence.SELF__}}[R]":
        pass

    @overload
    def map(self, selector: Callable[..., R], unpack="**") -> "{{sequence.SELF__}}[R]":
        pass

    @overload
    def map(self, selector: Callable[..., R], unpack="***") -> "{{sequence.SELF__}}[R]":
        pass

    def map(self, selector, unpack=""):
        return queryables.Map(self, selector, unpack)

    {% if query.is_pair %}

    @overload
    def select(self, item: Literal[0]) -> "Query[{{query.K}}]":
        ...

    @overload
    def select(self, item: Literal[1]) -> "Query[{{query.V}}]":
        ...

    {% endif %}

    @overload
    def select(self, field) -> "Query[Any]":
        ...

    @overload
    def select(self, field1, field2, *fields) -> "PairQuery":
        ...

    @overload
    def select(self, field, *fields) -> "Query[Tuple]":
        ...

    def select(self, *fields, attr: bool = False) -> "Query[Any]":
        return queryables.Select(self, *fields, attr=attr)

    def select_as_tuple(self, *fields, attr: bool = False) -> "Query[Tuple]":
        return queryables.SelectAsTuple(self, *fields, attr=attr)

    def select_as_dict(self, *fields, attr: bool = False, default=NoReturn) -> "Query[Dict]":
        return queryables.SelectAsDict(self, *fields, attr=attr, default=default)

    def reflect(self, mapping, default=NoReturn, attr: bool = False):
        return queryables.Reflect(self, mapping, attr=attr)

    def flat(self, selector: Callable[..., Iterable[R]] = None) -> "{{sequence.SELF__}}[R]":
        return queryables.Flat(self, selector)

    def traverse(self, selector: Callable[[{{query.T}}], Iterable[{{query.T}}]]) -> "{{sequence.SELF__}}[{{query.T}}]":
        return queryables.Traverse(self, selector)

    def pivot_unstack(self, default=None) -> "PairQuery[Any, List]":
        return queryables.PivotUnstack(self, default=default)

    def pivot_stack(self) -> "Query[Dict]":
        return queryables.PivotStack(self)

    def group_by(self, selector: Callable[[{{query.T}}], Tuple[K2, V2]] = lambda x: x) -> "{{pair.SELF__}}[K2, List[V2]]":
        return queryables.GroupBy(self, selector=selector)

    def chain(self, *iterables):
        return queryables.Chain(self, *iterables)

    def chunk(self, size: int) -> "Query[List[{{query.T}}]]":
        return queryables.Chunk(self, size=size)

    def tee(self, size: int):
        return queryables.Tee(self, size=size)

    def inner_join(self, right):
        return queryables.InnerJoin(self, right)

    def join(self, right, on: Callable[[Tuple[list, list]], Callable], select):
        return queryables.Join(self, right, on=on, select=select)

    def debug(self, breakpoint=lambda x: x, printer=print) -> "{{query.SELF_T}}":
        return queryables.Debug(self, breakpoint=breakpoint, printer=printer)

    def debug_path(self, selector_sync=lambda x: -10, selector_async=lambda x: 10) -> "{{query.SELF_T}}":
        return queryables.DebugPath(self, selector_sync, selector_async)

    def filter(self, predicate: Callable[[{{query.T}}], bool]) -> "{{query.SELF_T}}":
        return queryables.Filter(self, predicate)

    def filter_type(self, *types: Type[R]) -> "{{query.SELF_T}}":
        return queryables.FilterType(self, *types)

    @overload
    def filter_unique(self) -> "{{query.SELF_T}}":
        ...

    @overload
    def filter_unique(self, selector: Callable[[{{query.T}}], Tuple[K2, V2]]) -> "{{pair.SELF__}}[K2, V2]":
        ...

    @overload
    def filter_unique(self, selector: Callable[[{{query.T}}], R]) -> "{{sequence.SELF__}}[R]":
        ...

    def filter_unique(self, selector=None):
        return queryables.FilterUnique(self, selector=selector)

    def distinct(self, selector: Callable[[{{query.T}}], Any]) -> "{{query.SELF_T}}":
        return queryables.FilterUnique(self, selector=selector)

    def must(self, predicate: Callable[[{{query.T}}], bool], msg: str="") -> "{{query.SELF_T}}":
        return queryables.Must(self, predicate, msg)

    def must_type(self, type, *types: Type) -> "{{query.SELF_T}}":
        return queryables.MustType(self, type, *types)

    def must_unique(self, selector: Callable[[T], R] = None):
        return queryables.MustUnique(self, selector=selector)

    def take(self, count_or_range: Union[int, range]) -> "{{query.SELF_T}}":
        return queryables.Take(self, count_or_range)

    def take_while(self, predicate) -> "{{query.SELF_T}}":
        return queryables.TakeWhile(self, predicate)

    def skip(self, count_or_range: Union[int, range]) -> "{{query.SELF_T}}":
        return queryables.Skip(self, count_or_range)

    def take_page(self, page: int, size: int) -> "{{query.SELF_T}}":
        return queryables.TakePage(self, page=page, size=size)

    def order_by(self, *fields, desc: bool = False, attr: bool = False) -> "{{query.SELF_T}}":
        return queryables.OrderBy(self, *fields, desc=desc, attr=attr)

    def order_by_map(self, selector=None, *, desc: bool = False) -> "{{query.SELF_T}}":
        return queryables.OrderByMap(self, selector=selector, desc=desc)

    def order_by_reverse(self) -> "{{query.SELF_T}}":
        return queryables.OrderByReverse(self)

    def order_by_shuffle(self) -> "{{query.SELF_T}}":
        return queryables.OrderByShuffle(self)

    def sleep(self, seconds: float) -> "{{query.SELF_T}}":
        return queryables.Sleep(self, seconds)

    def zip(self):
        return queryables.Zip(self)

    def cartesian(self, *iterables) -> "Query[Tuple]":
        return queryables.Cartesian(self, *iterables)

{% endfor %}


if not TYPE_CHECKING:
    import types

    class Queries:
        pass

    from ._itertools import queryables

    classess = Queries()

    for cls in queryables.exports:
        baseclasses = (Query[T], cls)
        created = types.new_class(cls.__name__, baseclasses)

        setattr(classess, cls.__name__, created)

    queryables = classess


class QueryBase(Query[T], core.Query[T]):
    pass


class QueryAsync(Query[T], core.QueryAsync[T]):
    pass


class QueryNormal(Query[T], core.QueryNormal[T]):
    pass


class QueryDict(PairQuery[K, V], core.QueryDict[K, V]):
    def filter_keys(self, *keys) -> "PairQuery[K, V]":
        return queryables.FilterKeys(self, *keys)

    # @no_type_check
    def must_keys(self, *keys) -> "PairQuery[K, V]":
        return queryables.MustKeys(self, *keys, typ="map")

    def get(self, key) -> V:
        try:
            return self.source[key]
        except KeyError:
            raise NotFoundError(key)

    def get_or(self, key, default: R = None) -> Union[V, R]:
        try:
            return self.get(key)
        except NotFoundError:
            return default

    def get_or_raise(self, key, exc: Union[str, Exception]) -> V:
        undefined = object()
        result = self.get_or(key, undefined)
        if result is undefined:
            if isinstance(exc, str):
                raise Exception(exc)
            else:
                raise exc
        else:
            return result


class QuerySeq(Query[T], core.QuerySeq[T]):
    def filter_keys(self, *keys) -> "Query[T]":
        return queryables.FilterKeys(self, *keys)

    def must_keys(self, *keys) -> "Query[T]":
        return queryables.MustKeys(self, *keys, typ="seq")

    def get(self, key) -> T:
        try:
            return self.source[key]
        except IndexError:
            raise NotFoundError(key)

    def get_or(self, key, default: R = None) -> Union[T, R]:
        try:
            return self.get(key)
        except NotFoundError:
            return default

    def get_or_raise(self, key, exc: Union[str, Exception]) -> T:
        undefined = object()
        result = self.get_or(key, undefined)
        if result is undefined:
            if isinstance(exc, str):
                raise Exception(exc)
            else:
                raise exc
        else:
            return result


class PnqList(Query[T], List[T]):
    @property
    def source(self):
        return self

    @classmethod
    async def from_aiter(cls, aiter: AsyncIterable[T]) -> "PnqList[T]":
        result = cls()
        async for x in aiter:
            result.append(x)

        return result

    def filter_keys(self, *keys) -> "Query[T]":
        return queryables.FilterKeys(self, *keys)

    def must_keys(self, *keys) -> "Query[T]":
        return queryables.MustKeys(self, *keys, typ="seq")

    def get(self, key) -> T:
        try:
            return self.source[key]
        except IndexError:
            raise NotFoundError(key)

    def get_or(self, key, default: R = None) -> Union[T, R]:
        try:
            return self.get(key)
        except NotFoundError:
            return default

    def get_or_raise(self, key, exc: Union[str, Exception]) -> T:
        undefined = object()
        result = self.get_or(key, undefined)
        if result is undefined:
            if isinstance(exc, str):
                raise Exception(exc)
            else:
                raise exc
        else:
            return result


class QuerySet(Query[T], core.QuerySet[T]):
    def filter_keys(self, *keys) -> "Query[T]":
        return queryables.FilterKeys(self, *keys)

    def must_keys(self, *keys) -> "Query[T]":
        return queryables.MustKeys(self, *keys, typ="set")

    def get(self, key) -> T:
        if key in self.source:
            return key
        else:
            raise NotFoundError(key)

    def get_or(self, key, default: R = None) -> Union[T, R]:
        try:
            return self.get(key)
        except NotFoundError:
            return default

    def get_or_raise(self, key, exc: Union[str, Exception]) -> T:
        undefined = object()
        result = self.get_or(key, undefined)
        if result is undefined:
            if isinstance(exc, str):
                raise Exception(exc)
            else:
                raise exc
        else:
            return result

if TYPE_CHECKING:
    class QuerySeqPair(PairQuery[K, V], core.QuerySeq[Tuple[K, V]]):
        ...

    class QuerySetPair(PairQuery[K, V], core.QuerySeq[Tuple[K, V]]):
        ...

    class PnqListPair(PairQuery[K, V], PnqList[Tuple[K, V]]):
        def __init__(self, source: Iterable[Tuple[K, V]]):
            ...


class QueryBuilder(builder.Builder):
    QUERY_BOTH = QueryBase
    QUERY_ASYNC = QueryAsync
    QUERY_NORMAL = QueryNormal
    QUERY_SEQ = QuerySeq
    QUERY_DICT = QueryDict
    QUERY_SET = QuerySet


@overload
def query(source: Mapping[K, V]) -> QueryDict[K, V]:
    ...


@overload
def query(source: Set[Tuple[K, V]]) -> "QuerySetPair[K, V]":
    ...

@overload
def query(source: Set[T]) -> QuerySet[T]:
    ...

@overload
def query(source: Iterable[Tuple[K, V]]) -> "QuerySeqPair[K, V]":
    ...


@overload
def query(source: AsyncIterable[Tuple[K, V]]) -> "QuerySeqPair[K, V]":
    ...

@overload
def query(source: Generator[T, Any, Any]) -> QuerySeq[T]:
    ...

@overload
def query(source: AsyncGenerator[T, Any]) -> QuerySeq[T]:
    ...

@overload
def query(source: Iterable[T]) -> QuerySeq[T]:
    ...

@overload
def query(source: AsyncIterable[T]) -> QuerySeq[T]:
    ...

@overload
def query(source: Iterable[T]) -> QuerySeq[T]:
    ...

@overload
def query(source) -> "Query[Any]":
    ...


def query(source):
    return QueryBuilder.query(source)


run = QueryBuilder.run



import traceback
from datetime import datetime, timezone
from json import dumps as _dumps
from typing import Any, Dict, NamedTuple, Tuple, Union


def dumps(obj: Any) -> str:
    return _dumps(obj, ensure_ascii=False)


# TODO: is_cancelled はスペルが長くし、否定形で判定することが多いので嫌いだ。is_activeを採用したい。
class CancelToken:
    def __init__(self):
        self._is_active = True

    @property
    def is_cancelled(self):
        return not self._is_active

    @property
    def is_active(self):
        return self._is_active

    def cancel(self):
        self._is_active = False


class Request:
    args: Tuple
    kwargs: Dict


class Response(NamedTuple):
    func: Any
    args: Tuple
    kwargs: Dict
    err: Union[Exception, None]
    res: Any
    start: datetime
    end: datetime

    def result(self, timeout=None):
        if self.err:
            raise self.err
        else:
            return self.res

    @property
    def elapsed(self):
        return self.end - self.start

    @property
    def stack_trace(self):
        err = self.err
        if not err:
            return ""
        return "".join(
            traceback.format_exception(etype=type(err), value=err, tb=err.__traceback__)
        )

    def __str__(self):
        return str(self.to_dict())

    def to_dict(self, stack_trace: bool = True):
        """jsonに近いように辞書化します。kwargsとresultは解析されません。"""
        st = None
        err = None
        msg = None
        if self.err:
            err = self.err.__class__.__name__
            msg = str(self.err)
            if stack_trace:
                st = self.stack_trace

        return {
            "func": self.func.__name__,
            "args": self.args,
            "kwargs": self.kwargs,
            "result": self.res,
            "start": self.start.isoformat(),
            "end": self.end.isoformat(),
            "err": err,
            "msg": msg,
            "stack_trace": st,
        }

    def to_json(self, serializer=dumps):
        return serializer(self.to_dict())


class StopWatch:
    """コンテキスト内の処理時間を計測します。

    Args:

    * name: 任意の名前を付与できます

    Members:

    * name: 初期化時に付与した名前
    * start: コンテキストの開始時間（UTC）
    * end: コンテキストの完了時間（UTC）
    * elapsed: 開始時間と完了時間の差分秒数

    Usage:
    ```
    >>> with StopWatch("test") as result:
    >>>   [x for x in range(10000)]
    >>> print(result)
    {'name': 'test', 'start': '2021-09-13T14:10:04.780085+00:00', 'end': '2021-09-13T14:10:11.907716+00:00', 'elapsed': 7.127631}
    ```
    """

    name: str
    start: datetime
    end: datetime

    def __init__(self, name=""):
        self.name = name or ""
        self.start = None
        self.end = None

    def __enter__(self):
        if self.start:
            raise RuntimeError("StopWatch already started")
        start = datetime.utcnow()
        self.start = start.astimezone(timezone.utc)
        return self

    def __exit__(self, exc_value, exc_type, exc_tb):
        end = datetime.utcnow()
        self.end = end.astimezone(timezone.utc)

    def __str__(self):
        return str(self.to_dict())

    @property
    def elapsed(self):
        return (self.end - self.start).total_seconds()

    def to_dict(self):
        """計測データを辞書化します。日付データはisoformatで出力されます"""
        return {
            "name": self.name,
            "start": self.start.isoformat(),
            "end": self.end.isoformat(),
            "elapsed": self.elapsed,
        }
